from __future__ import annotations

from deprecated.sphinx import versionadded

from coredis.commands import CommandMixin
from coredis.typing import AnyStr

from .autocomplete import Autocomplete
from .filters import (
    BloomFilter,
    CountMinSketch,
    CuckooFilter,
    RedisBloom,
    TDigest,
    TopK,
)
from .graph import Graph
from .json import Json, RedisJSON
from .search import RediSearch, Search
from .timeseries import RedisTimeSeries, TimeSeries


class ModuleMixin(CommandMixin[AnyStr]):
    @property
    @versionadded(version="4.12.0")
    def json(self) -> Json[AnyStr]:
        """
        Property to access :class:`~coredis.modules.Json` commands.
        """
        return Json(self)

    @property
    @versionadded(version="4.12.0")
    def bf(self) -> BloomFilter[AnyStr]:
        """
        Property to access :class:`~coredis.modules.BloomFilter` commands.
        """
        return BloomFilter(self)

    @property
    @versionadded(version="4.12.0")
    def cf(self) -> CuckooFilter[AnyStr]:
        """
        Property to access :class:`~coredis.modules.CuckooFilter` commands.
        """
        return CuckooFilter(self)

    @property
    @versionadded(version="4.12.0")
    def cms(self) -> CountMinSketch[AnyStr]:
        """
        Property to access :class:`~coredis.modules.CountMinSketch` commands.
        """
        return CountMinSketch(self)

    @property
    @versionadded(version="4.12.0")
    def tdigest(self) -> TDigest[AnyStr]:
        """
        Property to access :class:`~coredis.modules.TDigest` commands.
        """
        return TDigest(self)

    @property
    @versionadded(version="4.12.0")
    def topk(self) -> TopK[AnyStr]:
        """
        Property to access :class:`~coredis.modules.TopK` commands.
        """
        return TopK(self)

    @property
    @versionadded(version="4.12.0")
    def timeseries(self) -> TimeSeries[AnyStr]:
        """
        Property to access :class:`~coredis.modules.TimeSeries` commands.
        """
        return TimeSeries(self)

    @property
    @versionadded(version="4.12.0")
    def search(self) -> Search[AnyStr]:
        """
        Property to access :class:`~coredis.modules.Search` commands.
        """
        return Search(self)

    @property
    @versionadded(version="4.12.0")
    def autocomplete(self) -> Autocomplete[AnyStr]:
        """
        Property to access :class:`~coredis.modules.Autocomplete` commands.
        """
        return Autocomplete(self)

    @property
    @versionadded(version="4.12.0")
    def graph(self) -> Graph[AnyStr]:
        """
        Property to access :class:`~coredis.modules.Graph` commands.
        """
        return Graph(self)


__all__ = [
    "RediSearch",
    "RedisBloom",
    "RedisJSON",
    "RedisTimeSeries",
    "Json",
    "BloomFilter",
    "CuckooFilter",
    "CountMinSketch",
    "TopK",
    "TDigest",
    "TimeSeries",
]



from __future__ import annotations

import asyncio
from types import TracebackType

from typing_extensions import runtime_checkable

from coredis.typing import (
    Awaitable,
    Callable,
    KeyT,
    Optional,
    Parameters,
    Protocol,
    R,
    ResponseType,
    StringT,
    Tuple,
    Type,
    TypeVar,
    ValueT,
)

T_co = TypeVar("T_co", covariant=True)


class AbstractExecutor(Protocol):
    async def execute_command(
        self,
        command: bytes,
        *args: ValueT,
        callback: Callable[..., R] = ...,
        **options: Optional[ValueT],
    ) -> R:
        ...


@runtime_checkable
class SupportsPipeline(Protocol):  # noqa
    async def pipeline(
        self,
        transaction: Optional[bool] = True,
        watches: Optional[Parameters[StringT]] = None,
    ) -> SupportsWatch:
        ...


@runtime_checkable
class SupportsScript(Protocol[T_co]):  # noqa
    async def evalsha(
        self,
        sha1: StringT,
        keys: Optional[Parameters[KeyT]] = ...,
        args: Optional[Parameters[ValueT]] = ...,
    ) -> ResponseType:
        ...

    async def evalsha_ro(
        self,
        sha1: StringT,
        keys: Optional[Parameters[KeyT]] = ...,
        args: Optional[Parameters[ValueT]] = ...,
    ) -> ResponseType:
        ...

    async def script_load(self, script: StringT) -> T_co:
        ...


@runtime_checkable
class SupportsWatch(Protocol):  # noqa
    async def __aenter__(self) -> SupportsWatch:
        ...

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> Awaitable[Optional[bool]]:
        ...

    async def watch(self, *keys: KeyT) -> bool:
        ...

    async def execute(self, raise_on_error: bool = True) -> Tuple[object, ...]:
        ...


@runtime_checkable
class ConnectionP(Protocol):
    decode_responses: bool
    encoding: str
    push_messages: asyncio.Queue[ResponseType]



from __future__ import annotations

import asyncio
import dataclasses
import functools
import inspect
import itertools
import os
import socket
import ssl
import time
import warnings
import weakref
from collections import defaultdict, deque
from typing import TYPE_CHECKING, Any, cast

import async_timeout

import coredis
from coredis._packer import Packer
from coredis._utils import nativestr
from coredis.exceptions import (
    AuthenticationRequiredError,
    ConnectionError,
    RedisError,
    TimeoutError,
    UnknownCommandError,
)
from coredis.parser import NotEnoughData, Parser
from coredis.tokens import PureToken
from coredis.typing import (
    Awaitable,
    Callable,
    ClassVar,
    Deque,
    Dict,
    List,
    Literal,
    Optional,
    ResponseType,
    Set,
    Tuple,
    TypeVar,
    Union,
    ValueT,
)

R = TypeVar("R")

if TYPE_CHECKING:
    from coredis.pool.nodemanager import ManagedNode


@dataclasses.dataclass
class Request:
    connection: weakref.ProxyType[Connection]
    command: bytes
    decode: bool
    encoding: Optional[str] = None
    raise_exceptions: bool = True
    future: asyncio.Future[ResponseType] = dataclasses.field(
        default_factory=lambda: asyncio.get_running_loop().create_future()
    )
    created_at: float = dataclasses.field(default_factory=lambda: time.time())

    def __post_init__(self) -> None:
        self.future.add_done_callback(self.cleanup)

    def cleanup(self, future: asyncio.Future[ResponseType]) -> None:
        if future.cancelled() and self.connection and self.connection.is_connected:
            self.connection.disconnect()

    def enforce_deadline(self, timeout: float) -> None:
        if not self.future.done():
            self.future.set_exception(
                TimeoutError(
                    f"command {nativestr(self.command)} timed out after {timeout} seconds"
                )
            )


@dataclasses.dataclass
class CommandInvocation:
    command: bytes
    args: Tuple[ValueT, ...]
    decode: Optional[bool]
    encoding: Optional[str]


class RedisSSLContext:
    context: Optional[ssl.SSLContext]

    def __init__(
        self,
        keyfile: Optional[str],
        certfile: Optional[str],
        cert_reqs: Optional[Union[str, ssl.VerifyMode]] = None,
        ca_certs: Optional[str] = None,
        check_hostname: Optional[bool] = None,
    ) -> None:
        self.keyfile = keyfile
        self.certfile = certfile
        self.check_hostname = check_hostname if check_hostname is not None else False
        if cert_reqs is None:
            self.cert_reqs = ssl.CERT_OPTIONAL
        elif isinstance(cert_reqs, str):
            CERT_REQS = {
                "none": ssl.CERT_NONE,
                "optional": ssl.CERT_OPTIONAL,
                "required": ssl.CERT_REQUIRED,
            }

            self.cert_reqs = CERT_REQS[cert_reqs]
        self.ca_certs = ca_certs
        self.context = None

    def get(self) -> ssl.SSLContext:
        if self.keyfile is None:
            self.context = ssl.create_default_context(cafile=self.ca_certs)
        else:
            self.context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
            self.context.verify_mode = self.cert_reqs
            self.context.check_hostname = self.check_hostname
            self.context.load_cert_chain(
                certfile=self.certfile, keyfile=self.keyfile  # type: ignore
            )
            if self.ca_certs:
                self.context.load_verify_locations(
                    **{
                        "capath"
                        if os.path.isdir(self.ca_certs)
                        else "cafile": self.ca_certs
                    }
                )
        assert self.context
        return self.context


class BaseConnection(asyncio.BaseProtocol):
    """
    Base connection class which implements
    :class:`asyncio.BaseProtocol` to interact
    with the underlying connection established
    with the redis server.
    """

    #: id for this connection as returned by the redis server
    client_id: Optional[int]
    #: Queue that collects any unread push message types
    push_messages: asyncio.Queue[ResponseType]
    #: client id that the redis server should send any redirected notifications to
    tracking_client_id: Optional[int]
    #: Whether the connection should use RESP or RESP3
    protocol_version: Literal[2, 3]

    description: ClassVar[str] = "BaseConnection"
    locator: ClassVar[str] = ""

    #: average response time of requests made on this connection
    average_response_time: float

    def __init__(
        self,
        stream_timeout: Optional[float] = None,
        encoding: str = "utf-8",
        decode_responses: bool = False,
        *,
        client_name: Optional[str] = None,
        protocol_version: Literal[2, 3] = 3,
        noreply: bool = False,
        noevict: bool = False,
        notouch: bool = False,
    ):
        self._stream_timeout = stream_timeout
        self.username: Optional[str] = None
        self.password: Optional[str] = ""
        self.db: Optional[int] = None
        self.pid: int = os.getpid()
        self._description_args: Callable[
            ..., Dict[str, Optional[Union[str, int]]]
        ] = lambda: dict()
        self._connect_callbacks: List[
            Union[
                Callable[[BaseConnection], Awaitable[None]],
                Callable[[BaseConnection], None],
            ]
        ] = list()
        self.encoding = encoding
        self.decode_responses = decode_responses
        self.protocol_version = protocol_version
        self.server_version: Optional[str] = None
        self.client_name = client_name
        self.client_id = None
        self.tracking_client_id = None

        self.last_active_at: float = time.time()
        self.last_request_processed_at: Optional[float] = None

        self._transport: Optional[asyncio.Transport] = None
        self._parser = Parser()
        self._read_flag = asyncio.Event()
        self.packer: Packer = Packer(self.encoding)
        self.push_messages: asyncio.Queue[ResponseType] = asyncio.Queue()

        self.noreply: bool = noreply
        self.noreply_set: bool = False

        self.noevict: bool = noevict
        self.notouch: bool = notouch

        self.needs_handshake: bool = True
        self._last_error: Optional[BaseException] = None
        self._connection_error: Optional[BaseException] = None

        self._requests: Deque[Request] = deque()

        self.average_response_time: float = 0
        self.requests_processed: int = 0
        self._write_ready: asyncio.Event = asyncio.Event()

    def __repr__(self) -> str:
        return self.describe(self._description_args())

    @classmethod
    def describe(cls, description_args: Dict[str, Any]) -> str:
        return cls.description.format_map(defaultdict(lambda: None, description_args))

    @property
    def location(self) -> str:
        return self.locator.format_map(
            defaultdict(lambda: None, self._description_args())
        )

    @property
    def estimated_time_to_idle(self) -> float:
        """
        Estimated time till the pending request queue of this connection
        has been cleared
        """
        return self.requests_pending * self.average_response_time

    def __del__(self) -> None:
        try:
            self.disconnect()
        except Exception:  # noqa
            pass

    @property
    def is_connected(self) -> bool:
        """
        Whether the connection is established and initial handshakes were
        performed without error
        """
        return self._transport is not None and self._connection_error is None

    @property
    def requests_pending(self) -> int:
        """
        Number of requests pending response on this connection
        """
        return len(self._requests)

    @property
    def lag(self) -> float:
        """
        Returns the amount of seconds since the last request was processed
        if there are still in flight requests pending on this connection
        """
        if not self._requests:
            return 0
        elif self.last_request_processed_at is None:
            return time.time()
        else:
            return time.time() - self.last_request_processed_at

    def register_connect_callback(
        self,
        callback: Union[
            Callable[[BaseConnection], None],
            Callable[[BaseConnection], Awaitable[None]],
        ],
    ) -> None:
        self._connect_callbacks.append(callback)

    def clear_connect_callbacks(self) -> None:
        self._connect_callbacks = list()

    async def can_read(self) -> bool:
        """Checks for data that can be read"""
        assert self._parser

        if not self.is_connected:
            await self.connect()

        return self._parser.can_read()

    async def connect(self) -> None:
        """
        Establish a connnection to the redis server
        and initiate any post connect callbacks
        """
        self._connection_error = None
        try:
            await self._connect()
        except (asyncio.CancelledError, RedisError) as err:
            self._connection_error = err
            raise
        except Exception as err:
            self._connection_error = err
            raise ConnectionError(str(err)) from err

        # run any user callbacks. right now the only internal callback
        # is for pubsub channel/pattern resubscription
        for callback in self._connect_callbacks:
            task = callback(self)
            if inspect.isawaitable(task):
                await task

    def connection_made(self, transport: asyncio.BaseTransport) -> None:
        """
        :meta private:
        """
        self._transport = cast(asyncio.Transport, transport)
        self._write_ready.set()

    def connection_lost(self, exc: Optional[BaseException]) -> None:
        """
        :meta private:
        """
        if exc:
            self._last_error = exc

        self.disconnect()

    def pause_writing(self) -> None:
        """
        :meta private:
        """
        self._write_ready.clear()

    def resume_writing(self) -> None:
        """
        :meta private:
        """
        self._write_ready.set()

    def data_received(self, data: bytes) -> None:
        """
        :meta private:
        """
        self._parser.feed(data)
        self._read_flag.set()
        if not self._requests:
            return

        request = self._requests.popleft()
        response = self._parser.get_response(request.decode, request.encoding)
        while not isinstance(
            response,
            NotEnoughData,
        ):
            if not (request.future.cancelled() or request.future.done()):
                if request.raise_exceptions and isinstance(response, RedisError):
                    request.future.set_exception(response)
                else:
                    request.future.set_result(response)

            self.last_request_processed_at = time.time()
            self.requests_processed += 1
            response_time = time.time() - request.created_at

            self.average_response_time = (
                (self.average_response_time * (self.requests_processed - 1))
                + response_time
            ) / self.requests_processed

            try:
                request = self._requests.popleft()
            except IndexError:
                return

            response = self._parser.get_response(request.decode, request.encoding)

        # In case the first request pulled from the queue doesn't have enough data
        # to process, put it back to the start of the queue for the next iteration
        if request:
            self._requests.appendleft(request)

    def eof_received(self) -> None:
        """
        :meta private:
        """
        self.disconnect()

    async def _connect(self) -> None:
        raise NotImplementedError

    async def update_tracking_client(
        self, enabled: bool, client_id: Optional[int] = None
    ) -> bool:
        """
        Associate this connection to :paramref:`client_id` to
        relay any tracking notifications to.
        """
        try:
            params: List[ValueT] = (
                [b"ON", b"REDIRECT", client_id]
                if (enabled and client_id is not None)
                else [b"OFF"]
            )

            if (
                await (
                    await self.create_request(b"CLIENT TRACKING", *params, decode=False)
                )
                != b"OK"
            ):
                raise ConnectionError("Unable to toggle client tracking")
            self.tracking_client_id = client_id
            return True
        except UnknownCommandError:  # noqa
            raise
        except Exception:  # noqa
            return False

    async def try_legacy_auth(self) -> None:
        if not self.password:
            return
        params = [self.password]
        if self.username:
            params.insert(0, self.username)
        await (await self.create_request(b"AUTH", *params, decode=False))

    async def perform_handshake(self) -> None:
        if not self.needs_handshake:
            return

        hello_command_args: List[Union[int, str, bytes]] = [self.protocol_version]
        if self.username or self.password:
            hello_command_args.extend(
                ["AUTH", self.username or b"default", self.password or b""]
            )
        try:
            hello_resp = await (
                await self.create_request(b"HELLO", *hello_command_args, decode=False)
            )
            assert isinstance(hello_resp, (list, dict))
            if self.protocol_version == 3:
                resp3 = cast(Dict[bytes, ValueT], hello_resp)
                assert resp3[b"proto"] == 3
                self.server_version = nativestr(resp3[b"version"])
                self.client_id = int(resp3[b"id"])
            else:
                resp = cast(List[ValueT], hello_resp)
                self.server_version = nativestr(resp[3])
                self.client_id = int(resp[7])
            # TODO: change this to 7.2 once it is officially released?
            if self.server_version > "7.1":
                await asyncio.gather(
                    await self.create_request(
                        b"CLIENT SETINFO",
                        b"LIB-NAME",
                        b"coredis",
                    ),
                    await self.create_request(
                        b"CLIENT SETINFO",
                        b"LIB-VER",
                        coredis.__version__,
                    ),
                )
            self.needs_handshake = False
        except AuthenticationRequiredError:
            await self.try_legacy_auth()
            self.server_version = None
            self.client_id = None
        except UnknownCommandError:  # noqa
            # This should only happen for redis servers < 6 or forks of redis
            # that are not > 6 compliant.
            warning = (
                "The server responded with no support for the `HELLO` command"
                " and therefore a handshake could not be performed"
            )
            if self.protocol_version == 3:
                raise ConnectionError(
                    "Unable to use RESP3 due to missing `HELLO` implementation "
                    "the server. Use `protocol_version=2` when constructing the client."
                )
            else:
                warnings.warn(warning, category=UserWarning)
                await self.try_legacy_auth()
            self.needs_handshake = False

    async def on_connect(self) -> None:
        self._parser.on_connect(self)
        await self.perform_handshake()

        if self.db:
            if (
                await (await self.create_request(b"SELECT", self.db, decode=False))
                != b"OK"
            ):
                raise ConnectionError(f"Invalid Database {self.db}")

        if self.client_name is not None:
            if (
                await (
                    await self.create_request(
                        b"CLIENT SETNAME", self.client_name, decode=False
                    )
                )
                != b"OK"
            ):
                raise ConnectionError(f"Failed to set client name: {self.client_name}")

        if self.noevict:
            await (await self.create_request(b"CLIENT NO-EVICT", b"ON"))

        if self.notouch:
            await (await self.create_request(b"CLIENT NO-TOUCH", b"ON"))

        if self.noreply:
            await (await self.create_request(b"CLIENT REPLY", b"OFF", noreply=True))
            self.noreply_set = True

        self.last_active_at = time.time()

    async def fetch_push_message(
        self,
        decode: Optional[ValueT] = None,
        push_message_types: Optional[Set[bytes]] = None,
        block: Optional[bool] = False,
    ) -> ResponseType:
        """
        Read the next pending response
        """
        if not self.is_connected:
            await self.connect()

        if len(self._requests) > 0:
            raise ConnectionError(
                f"Invalid request for push messages. {len(self._requests)} requests still pending"
            )

        message = self._parser.get_response(
            bool(decode) if decode is not None else self.decode_responses,
            self.encoding,
            push_message_types,
        )
        while isinstance(
            message,
            NotEnoughData,
        ):
            self._read_flag.clear()
            try:
                async with async_timeout.timeout(
                    self._stream_timeout if not block else None
                ):
                    await self._read_flag.wait()
            except asyncio.TimeoutError:
                raise TimeoutError
            message = self._parser.get_response(
                bool(decode) if decode is not None else self.decode_responses,
                self.encoding,
                push_message_types,
            )
        return message

    async def _send_packed_command(
        self, command: List[bytes], timeout: Optional[float] = None
    ) -> None:
        """
        Sends an already packed command to the Redis server
        """

        assert self._transport
        try:
            async with async_timeout.timeout(timeout):
                await self._write_ready.wait()
        except asyncio.TimeoutError:
            if self._transport:
                self.disconnect()
            raise TimeoutError(
                f"Unable to write after waiting for socket for {timeout} seconds"
            )
        self._transport.writelines(command)

    async def send_command(
        self,
        command: bytes,
        *args: ValueT,
    ) -> None:
        """
        Send a command to the redis server
        """

        if not self.is_connected:
            await self.connect()

        await self._send_packed_command(self.packer.pack_command(command, *args))

        self.last_active_at = time.time()

    async def create_request(
        self,
        command: bytes,
        *args: ValueT,
        noreply: Optional[bool] = None,
        decode: Optional[ValueT] = None,
        encoding: Optional[str] = None,
        raise_exceptions: bool = True,
        timeout: Optional[float] = None,
    ) -> asyncio.Future[ResponseType]:
        """
        Send a command to the redis server
        """
        from coredis.commands.constants import CommandName

        if not self.is_connected:
            await self.connect()

        cmd_list = []
        request_timeout: Optional[float] = timeout or self._stream_timeout
        if self.is_connected and noreply and not self.noreply:
            cmd_list = self.packer.pack_command(
                CommandName.CLIENT_REPLY, PureToken.SKIP
            )
        cmd_list.extend(self.packer.pack_command(command, *args))
        await self._send_packed_command(cmd_list, timeout=request_timeout)

        self.last_active_at = time.time()

        if not (self.noreply_set or noreply):
            request = Request(
                weakref.proxy(self),
                command,
                bool(decode) if decode is not None else self.decode_responses,
                encoding or self.encoding,
                raise_exceptions,
            )
            self._requests.append(request)
            if request_timeout is not None:
                asyncio.get_running_loop().call_later(
                    request_timeout,
                    functools.partial(
                        request.enforce_deadline,
                        request_timeout,
                    ),
                )
            return request.future
        else:
            none: asyncio.Future[ResponseType] = asyncio.Future()
            none.set_result(None)
            return none

    async def create_requests(
        self,
        commands: List[CommandInvocation],
        raise_exceptions: bool = True,
        timeout: Optional[float] = None,
    ) -> List[asyncio.Future[ResponseType]]:
        """
        Send multiple commands to the redis server
        """

        if not self.is_connected:
            await self.connect()

        request_timeout: Optional[float] = timeout or self._stream_timeout

        await self._send_packed_command(
            self.packer.pack_commands(
                list(itertools.chain((cmd.command, *cmd.args) for cmd in commands))
            ),
            timeout=request_timeout,
        )

        self.last_active_at = time.time()
        requests: List[asyncio.Future[ResponseType]] = []
        for cmd in commands:
            request = Request(
                weakref.proxy(self),
                cmd.command,
                bool(cmd.decode) if cmd.decode is not None else self.decode_responses,
                cmd.encoding or self.encoding,
                raise_exceptions,
            )
            self._requests.append(request)
            if request_timeout is not None:
                asyncio.get_running_loop().call_later(
                    request_timeout,
                    functools.partial(request.enforce_deadline, request_timeout),
                )
            requests.append(request.future)
        return requests

    def disconnect(self) -> None:
        """
        Disconnect from the Redis server
        """
        self.needs_handshake = True
        self.noreply_set = False
        self._parser.on_disconnect()
        if self._transport:
            try:
                self._transport.close()
            # Raised if event loop is already closed.
            except RuntimeError:  # noqa
                pass

        while True:
            try:
                request = self._requests.popleft()
                if not request.future.done():
                    request.future.set_exception(
                        self._last_error or ConnectionError("Connection lost")
                    )
            except IndexError:
                break
        self._transport = None


class Connection(BaseConnection):
    description: ClassVar[str] = "Connection<host={host},port={port},db={db}>"
    locator: ClassVar[str] = "host={host},port={port}"

    def __init__(
        self,
        host: str = "127.0.0.1",
        port: int = 6379,
        username: Optional[str] = None,
        password: Optional[str] = None,
        db: Optional[int] = 0,
        stream_timeout: Optional[float] = None,
        connect_timeout: Optional[float] = None,
        ssl_context: Optional[ssl.SSLContext] = None,
        encoding: str = "utf-8",
        decode_responses: bool = False,
        socket_keepalive: Optional[bool] = None,
        socket_keepalive_options: Optional[Dict[int, Union[int, bytes]]] = None,
        *,
        client_name: Optional[str] = None,
        protocol_version: Literal[2, 3] = 3,
        noreply: bool = False,
        noevict: bool = False,
        notouch: bool = False,
    ):
        super().__init__(
            stream_timeout,
            encoding,
            decode_responses,
            client_name=client_name,
            protocol_version=protocol_version,
            noreply=noreply,
            noevict=noevict,
            notouch=notouch,
        )
        self.host = host
        self.port = port
        self.username: Optional[str] = username
        self.password: Optional[str] = password
        self.db: Optional[int] = db
        self.ssl_context = ssl_context
        self._connect_timeout = connect_timeout
        self._description_args: Callable[
            ..., Dict[str, Optional[Union[str, int]]]
        ] = lambda: {
            "host": self.host,
            "port": self.port,
            "db": self.db,
        }
        self.socket_keepalive = socket_keepalive
        self.socket_keepalive_options: Dict[int, Union[int, bytes]] = (
            socket_keepalive_options or {}
        )

    async def _connect(self) -> None:
        if self.ssl_context:
            connection = asyncio.get_running_loop().create_connection(
                lambda: self, host=self.host, port=self.port, ssl=self.ssl_context
            )
        else:
            connection = asyncio.get_running_loop().create_connection(
                lambda: self, host=self.host, port=self.port
            )

        try:
            async with async_timeout.timeout(self._connect_timeout):
                transport, _ = await connection
        except asyncio.TimeoutError:
            raise ConnectionError(
                f"Unable to establish a connection within {self._connect_timeout} seconds"
            )
        sock = transport.get_extra_info("socket")
        if sock is not None:
            try:
                # TCP_KEEPALIVE
                if self.socket_keepalive:
                    sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)

                    for k, v in self.socket_keepalive_options.items():
                        sock.setsockopt(socket.SOL_TCP, k, v)
            except (OSError, TypeError):
                # `socket_keepalive_options` might contain invalid options
                # causing an error
                transport.close()
                raise
        await self.on_connect()


class UnixDomainSocketConnection(BaseConnection):
    description: ClassVar[str] = "UnixDomainSocketConnection<path={path},db={db}>"
    locator: ClassVar[str] = "path={path}"

    def __init__(
        self,
        path: str = "",
        username: Optional[str] = None,
        password: Optional[str] = None,
        db: int = 0,
        stream_timeout: Optional[float] = None,
        connect_timeout: Optional[float] = None,
        encoding: str = "utf-8",
        decode_responses: bool = False,
        *,
        client_name: Optional[str] = None,
        protocol_version: Literal[2, 3] = 3,
        **_: ValueT,
    ) -> None:
        super().__init__(
            stream_timeout,
            encoding,
            decode_responses,
            client_name=client_name,
            protocol_version=protocol_version,
        )
        self.path = path
        self.db = db
        self.username = username
        self.password = password
        self._connect_timeout = connect_timeout
        self._description_args = lambda: {"path": self.path, "db": self.db}

    async def _connect(self) -> None:
        async with async_timeout.timeout(self._connect_timeout):
            await asyncio.get_running_loop().create_unix_connection(
                lambda: self, path=self.path
            )

        await self.on_connect()


class ClusterConnection(Connection):
    "Manages TCP communication to and from a Redis server"
    description: ClassVar[str] = "ClusterConnection<host={host},port={port}>"
    locator: ClassVar[str] = "host={host},port={port}"
    node: "ManagedNode"

    def __init__(
        self,
        host: str = "127.0.0.1",
        port: int = 6379,
        username: Optional[str] = None,
        password: Optional[str] = None,
        db: Optional[int] = 0,
        stream_timeout: Optional[float] = None,
        connect_timeout: Optional[float] = None,
        ssl_context: Optional[ssl.SSLContext] = None,
        encoding: str = "utf-8",
        decode_responses: bool = False,
        socket_keepalive: Optional[bool] = None,
        socket_keepalive_options: Optional[Dict[int, Union[int, bytes]]] = None,
        *,
        client_name: Optional[str] = None,
        protocol_version: Literal[2, 3] = 3,
        read_from_replicas: bool = False,
        noreply: bool = False,
        noevict: bool = False,
        notouch: bool = False,
    ) -> None:
        self.read_from_replicas = read_from_replicas
        super().__init__(
            host=host,
            port=port,
            username=username,
            password=password,
            db=db,
            stream_timeout=stream_timeout,
            connect_timeout=connect_timeout,
            ssl_context=ssl_context,
            encoding=encoding,
            decode_responses=decode_responses,
            socket_keepalive=socket_keepalive,
            socket_keepalive_options=socket_keepalive_options,
            client_name=client_name,
            protocol_version=protocol_version,
            noreply=noreply,
            noevict=noevict,
            notouch=notouch,
        )

    async def on_connect(self) -> None:
        """
        Initialize the connection, authenticate and select a database and send
        `READONLY` if `read_from_replicas` is set during initialization.

        :meta private:
        """

        await super().on_connect()
        if self.read_from_replicas:
            assert (
                await (await self.create_request(b"READONLY", decode=False))
            ) == b"OK"



"""
coredis.response.callbacks
--------------------------
"""

# pyright: reportUnnecessaryIsInstance=false
from __future__ import annotations

import datetime
import itertools
from abc import ABC, ABCMeta, abstractmethod
from typing import TYPE_CHECKING, Any, cast

from coredis._utils import b
from coredis.exceptions import ClusterResponseError, ResponseError
from coredis.typing import (
    AnyStr,
    Callable,
    Dict,
    Generic,
    Hashable,
    Iterable,
    List,
    Literal,
    Mapping,
    Optional,
    ParamSpec,
    Protocol,
    ResponsePrimitive,
    ResponseType,
    Sequence,
    Set,
    StringT,
    Tuple,
    Type,
    TypeVar,
    Union,
    ValueT,
    add_runtime_checks,
    runtime_checkable,
)

R = TypeVar("R")
S = TypeVar("S")
T = TypeVar("T")
P = ParamSpec("P")
CR_co = TypeVar("CR_co", covariant=True)
CK_co = TypeVar("CK_co", covariant=True)

RESP = TypeVar("RESP")
RESP3 = TypeVar("RESP3")

if TYPE_CHECKING:
    from coredis.client import Client


class ResponseCallbackMeta(ABCMeta):
    def __new__(
        cls, name: str, bases: Tuple[type, ...], namespace: Dict[str, str]
    ) -> ResponseCallbackMeta:
        kls = super().__new__(cls, name, bases, namespace)
        setattr(kls, "transform", add_runtime_checks(getattr(kls, "transform")))
        setattr(kls, "transform_3", add_runtime_checks(getattr(kls, "transform_3")))
        return kls


class ClusterCallbackMeta(ABCMeta):
    def __new__(
        cls, name: str, bases: Tuple[type, ...], namespace: Dict[str, str]
    ) -> ClusterCallbackMeta:
        kls = super().__new__(cls, name, bases, namespace)
        setattr(kls, "combine", add_runtime_checks(getattr(kls, "combine")))
        setattr(kls, "combine_3", add_runtime_checks(getattr(kls, "combine_3")))
        return kls


class ResponseCallback(ABC, Generic[RESP, RESP3, R], metaclass=ResponseCallbackMeta):
    version: Literal[2, 3]

    def __call__(
        self,
        response: Union[RESP, RESP3, ResponseError],
        version: Literal[2, 3] = 2,
        **options: Optional[ValueT],
    ) -> R:
        self.version = version
        if isinstance(response, ResponseError):
            exc_to_response = self.handle_exception(response)
            if exc_to_response:
                return exc_to_response
        if version == 3:
            return self.transform_3(cast(RESP3, response), **options)
        return self.transform(cast(RESP, response), **options)

    @abstractmethod
    def transform(self, response: RESP, **options: Optional[ValueT]) -> R:
        pass

    def transform_3(self, response: RESP3, **options: Optional[ValueT]) -> R:
        return self.transform(cast(RESP, response), **options)

    def handle_exception(self, exc: BaseException) -> Optional[R]:
        return exc  # type: ignore


@runtime_checkable
class AsyncPreProcessingCallback(Protocol):
    async def pre_process(
        self, client: Client[Any], response: ResponseType, **options: Optional[ValueT]
    ) -> None:
        ...


class NoopCallback(ResponseCallback[R, R, R]):
    def transform(self, response: R, **kwargs: Optional[ValueT]) -> R:
        return response


class ClusterMultiNodeCallback(ABC, Generic[R], metaclass=ClusterCallbackMeta):
    def __call__(
        self, responses: Mapping[str, R], version: int = 2, **kwargs: Optional[ValueT]
    ) -> R:
        if version == 3:
            return self.combine_3(responses, **kwargs)
        return self.combine(responses, **kwargs)

    @property
    @abstractmethod
    def response_policy(self) -> str:
        ...

    @abstractmethod
    def combine(self, responses: Mapping[str, R], **kwargs: Optional[ValueT]) -> R:
        pass

    def combine_3(self, responses: Mapping[str, R], **kwargs: Optional[ValueT]) -> R:
        return self.combine(responses, **kwargs)

    @classmethod
    def raise_any(cls, values: Iterable[R]) -> None:
        for value in values:
            if isinstance(value, BaseException):
                raise value


class ClusterBoolCombine(ClusterMultiNodeCallback[bool]):
    def __init__(self, any: bool = False):
        self.any = any

    def combine(
        self, responses: Mapping[str, bool], **kwargs: Optional[ValueT]
    ) -> bool:
        values = tuple(responses.values())
        self.raise_any(values)
        assert (isinstance(value, bool) for value in values)
        return any(values) if self.any else all(values)

    @property
    def response_policy(self) -> str:
        return (
            "success if any shards responded ``True``"
            if self.any
            else "success if all shards responded ``True``"
        )


class ClusterAlignedBoolsCombine(ClusterMultiNodeCallback[Tuple[bool, ...]]):
    def combine(
        self, responses: Mapping[str, Tuple[bool, ...]], **kwargs: Optional[ValueT]
    ) -> Tuple[bool, ...]:
        return tuple(all(k) for k in zip(*responses.values()))

    @property
    def response_policy(self) -> str:
        return "the logical AND of all responses"


class ClusterEnsureConsistent(ClusterMultiNodeCallback[Optional[R]]):
    def __init__(self, ensure_consistent: bool = True):
        self.ensure_consistent = ensure_consistent

    def combine(
        self, responses: Mapping[str, Optional[R]], **kwargs: Optional[ValueT]
    ) -> Optional[R]:
        values = tuple(responses.values())
        self.raise_any(values)
        if self.ensure_consistent and len(set(values)) != 1:
            raise ClusterResponseError("Inconsistent response from cluster nodes")
        elif values:
            return values[0]
        return None

    @property
    def response_policy(self) -> str:
        return (
            "the response from any shard if all responses are consistent"
            if self.ensure_consistent
            else "first response"
        )


class ClusterFirstNonException(ClusterMultiNodeCallback[Optional[R]]):
    def combine(
        self, responses: Mapping[str, Optional[R]], **kwargs: Optional[ValueT]
    ) -> Optional[R]:
        for r in responses.values():
            if not isinstance(r, BaseException):
                return r
        for r in responses.values():
            if isinstance(r, BaseException):
                raise r

    @property
    def response_policy(self) -> str:
        return "the first response that is not an error"


class ClusterMergeSets(ClusterMultiNodeCallback[Set[R]]):
    def combine(
        self, responses: Mapping[str, Set[R]], **kwargs: Optional[ValueT]
    ) -> Set[R]:
        self.raise_any(responses.values())
        return set(itertools.chain(*responses.values()))

    @property
    def response_policy(self) -> str:
        return "the union of the results"


class ClusterSum(ClusterMultiNodeCallback[int]):
    def combine(self, responses: Mapping[str, int], **kwargs: Optional[ValueT]) -> int:
        self.raise_any(responses.values())
        return sum(responses.values())

    @property
    def response_policy(self) -> str:
        return "the sum of results"


class ClusterMergeMapping(ClusterMultiNodeCallback[Dict[CK_co, CR_co]]):
    def __init__(self, value_combine: Callable[[Iterable[CR_co]], CR_co]) -> None:
        self.value_combine = value_combine

    def combine(
        self, responses: Mapping[str, Dict[CK_co, CR_co]], **kwargs: Optional[ValueT]
    ) -> Dict[CK_co, CR_co]:
        self.raise_any(responses.values())
        response: Dict[CK_co, CR_co] = {}
        for key in set(itertools.chain(*responses.values())):
            values = list(responses[n][key] for n in responses if key in responses[n])
            response[key] = self.value_combine(values)
        return response

    @property
    def response_policy(self) -> str:
        return "the merged mapping"


class ClusterConcatenateTuples(ClusterMultiNodeCallback[Tuple[R, ...]]):
    def combine(
        self, responses: Mapping[str, Tuple[R, ...]], **kwargs: Optional[ValueT]
    ) -> Tuple[R, ...]:
        self.raise_any(responses.values())
        return tuple(itertools.chain(*responses.values()))

    @property
    def response_policy(self) -> str:
        return "the concatenations of the results"


class SimpleStringCallback(
    ResponseCallback[Optional[StringT], Optional[StringT], bool]
):
    def __init__(
        self,
        raise_on_error: Optional[Type[Exception]] = None,
        prefix_match: bool = False,
        ok_values: Set[str] = {"OK"},
    ):
        self.raise_on_error = raise_on_error
        self.prefix_match = prefix_match
        self.ok_values = {b(v) for v in ok_values}

    def transform(
        self, response: Optional[StringT], **options: Optional[ValueT]
    ) -> bool:
        if response:
            if not self.prefix_match:
                success = b(response) in self.ok_values
            else:
                success = any(b(response).startswith(ok) for ok in self.ok_values)
        else:
            success = False
        if not success and self.raise_on_error:
            raise self.raise_on_error(response)
        return success


class IntCallback(ResponseCallback[int, int, int]):
    def transform(
        self, response: ResponsePrimitive, **options: Optional[ValueT]
    ) -> int:
        if isinstance(response, int):
            return response
        raise ValueError(f"Unable to map {response!r} to int")


class AnyStrCallback(ResponseCallback[StringT, StringT, AnyStr]):
    def transform(self, response: StringT, **options: Optional[ValueT]) -> AnyStr:
        if isinstance(response, (bytes, str)):
            return cast(AnyStr, response)

        raise ValueError(f"Unable to map {response!r} to AnyStr")


class FloatCallback(
    ResponseCallback[Union[StringT, int, float], Union[StringT, int, float], float]
):
    def transform(self, response: ResponseType, **options: Optional[ValueT]) -> float:
        if isinstance(response, float):
            return response
        if isinstance(response, (int, bytes, str)):
            return float(response)

        raise ValueError(f"Unable to map {response} to float")


class BoolCallback(ResponseCallback[Union[int, bool], Union[int, bool], bool]):
    def transform(self, response: ResponseType, **options: Optional[ValueT]) -> bool:
        if isinstance(response, bool):
            return response
        return bool(response)


class SimpleStringOrIntCallback(ResponseCallback[ValueT, ValueT, Union[bool, int]]):
    def transform(
        self, response: ValueT, **options: Optional[ValueT]
    ) -> Union[bool, int]:
        if isinstance(response, (int, bool)):
            return response
        elif isinstance(response, (str, bytes)):
            return SimpleStringCallback()(response)
        raise ValueError(f"Unable to map {response!r} to bool")


class TupleCallback(
    ResponseCallback[List[ResponseType], List[ResponseType], Tuple[CR_co, ...]]
):
    def transform(
        self, response: ResponseType, **options: Optional[ValueT]
    ) -> Tuple[CR_co, ...]:
        if isinstance(response, List):
            return cast(Tuple[CR_co, ...], tuple(response))
        raise ValueError(f"Unable to map {response!r} to tuple")


class MixedTupleCallback(
    ResponseCallback[List[ResponseType], List[ResponseType], Tuple[R, S]]
):
    def transform(
        self, response: ResponseType, **options: Optional[ValueT]
    ) -> Tuple[R, S]:
        if isinstance(response, List):
            return cast(Tuple[R, S], tuple(response))
        raise ValueError(f"Unable to map {response!r} to tuple")


class ListCallback(
    ResponseCallback[List[ResponseType], List[ResponseType], List[CR_co]]
):
    def transform(
        self, response: List[ResponseType], **options: Optional[ValueT]
    ) -> List[CR_co]:
        return cast(List[CR_co], response)


class DateTimeCallback(
    ResponseCallback[Union[int, float], Union[int, float], datetime.datetime]
):
    def transform(
        self, response: Union[int, float], **kwargs: Optional[ValueT]
    ) -> datetime.datetime:
        ts = float(response) if not isinstance(response, float) else response
        if kwargs.get("unit") == "milliseconds":
            ts = ts / 1000.0
        return datetime.datetime.fromtimestamp(ts)


class DictCallback(
    ResponseCallback[
        Union[Sequence[ResponseType], Dict[ResponsePrimitive, ResponseType]],
        Union[Sequence[ResponseType], Dict[ResponsePrimitive, ResponseType]],
        Dict[CK_co, CR_co],
    ]
):
    def __init__(
        self,
        flat: bool = True,
        recursive: Optional[List[str]] = None,
    ):
        self.flat = flat
        self.recursive = recursive or []

    def transform(
        self,
        response: Union[Sequence[ResponseType], Dict[ResponsePrimitive, ResponseType]],
        **options: Optional[ValueT],
    ) -> Dict[CK_co, CR_co]:
        if isinstance(response, list):
            if self.flat:
                if self.recursive:
                    return cast(
                        Dict[CK_co, CR_co], self.recursive_transformer(response)
                    )
                else:
                    it = iter(response)
                    return cast(Dict[CK_co, CR_co], dict(zip(it, it)))
            else:
                return dict(r for r in response)
        raise ValueError(f"Unable to map {response!r} to mapping")

    def transform_3(
        self,
        response: Union[Sequence[ResponseType], Dict[ResponsePrimitive, ResponseType]],
        **options: Optional[ValueT],
    ) -> Dict[CK_co, CR_co]:
        if isinstance(response, Dict):
            return cast(Dict[CK_co, CR_co], response)
        return self.transform(response, **options)

    def recursive_transformer(
        self, item: Union[Sequence[ResponseType], Dict[ResponsePrimitive, ResponseType]]
    ) -> Union[
        Dict[CK_co, CR_co],
        List[CK_co],
        List[CR_co],
        Tuple[CK_co, ...],
        Tuple[CR_co, ...],
    ]:
        if isinstance(item, (list, tuple)):
            if len(item) % 2 == 0 and all(isinstance(k, Hashable) for k in item[::2]):
                dct = []
                for i in range(0, len(item), 2):
                    key, value = item[i], item[i + 1]
                    value = (
                        self.recursive_transformer(value)
                        if (key in self.recursive)
                        else value
                    )
                    dct.append((key, value))
                return dict(dct)
            else:
                caster = list if isinstance(item, list) else tuple
                return caster(self.recursive_transformer(i) for i in item)
        else:
            return item


class SetCallback(
    ResponseCallback[
        List[ResponsePrimitive],
        Set[ResponsePrimitive],
        Set[CR_co],
    ]
):
    def transform(
        self,
        response: Union[List[ResponsePrimitive], Set[ResponsePrimitive]],
        **options: Optional[ValueT],
    ) -> Set[CR_co]:
        if isinstance(response, list):
            return cast(Set[CR_co], set(response))
        raise ValueError(f"Unable to map {response} to set")

    def transform_3(
        self,
        response: Union[List[ResponsePrimitive], Set[ResponsePrimitive]],
        **options: Optional[ValueT],
    ) -> Set[CR_co]:
        if isinstance(response, set):
            return cast(Set[CR_co], response)
        else:
            return self.transform(response)


class OneOrManyCallback(
    ResponseCallback[
        Optional[Union[CR_co, List[Optional[CR_co]]]],
        Optional[Union[CR_co, List[Optional[CR_co]]]],
        Optional[Union[CR_co, List[Optional[CR_co]]]],
    ]
):
    def transform(
        self,
        response: Optional[Union[CR_co, List[Optional[CR_co]]]],
        **options: Optional[ValueT],
    ) -> Optional[Union[CR_co, List[Optional[CR_co]]]]:
        return response


class BoolsCallback(ResponseCallback[ResponseType, ResponseType, Tuple[bool, ...]]):
    def transform(
        self, response: ResponseType, **options: Optional[ValueT]
    ) -> Tuple[bool, ...]:
        if isinstance(response, List):
            return tuple(BoolCallback()(r) for r in response)
        return ()


class FloatsCallback(ResponseCallback[ResponseType, ResponseType, Tuple[float, ...]]):
    def transform(
        self, response: ResponseType, **options: Optional[ValueT]
    ) -> Tuple[float, ...]:
        if isinstance(response, List):
            return tuple(FloatCallback()(r) for r in response)
        return ()


class OptionalFloatCallback(
    ResponseCallback[
        Optional[Union[StringT, int, float]],
        Optional[Union[StringT, int, float]],
        Optional[float],
    ]
):
    def transform(
        self,
        response: Optional[Union[StringT, int, float]],
        **options: Optional[ValueT],
    ) -> Optional[float]:
        if response is None:
            return None
        return FloatCallback()(response)


class OptionalIntCallback(
    ResponseCallback[Optional[int], Optional[int], Optional[int]]
):
    def transform(
        self, response: Optional[int], **options: Optional[ValueT]
    ) -> Optional[int]:
        if response is None:
            return None
        if isinstance(response, int):
            return response
        raise ValueError(f"Unable to map {response} to int")


class OptionalAnyStrCallback(
    ResponseCallback[
        Optional[StringT],
        Optional[AnyStr],
        Optional[AnyStr],
    ]
):
    def transform(
        self, response: Optional[StringT], **options: Optional[ValueT]
    ) -> Optional[AnyStr]:
        if response is None:
            return None
        if isinstance(response, (bytes, str)):
            return cast(AnyStr, response)
        raise ValueError(f"Unable to map {response} to AnyStr")


class OptionalListCallback(
    ResponseCallback[List[ResponseType], List[ResponseType], Optional[List[CR_co]]]
):
    def transform(
        self, response: ResponseType, **options: Optional[ValueT]
    ) -> Optional[List[CR_co]]:
        return cast(List[CR_co], response)


class FirstValueCallback(ResponseCallback[List[CR_co], List[CR_co], CR_co]):
    def transform(self, response: List[CR_co], **options: Optional[ValueT]) -> CR_co:
        if response:
            return response[0]
        else:
            raise ValueError("Empty response list")



from __future__ import annotations

import asyncio
import threading
from asyncio import AbstractEventLoop, CancelledError
from concurrent.futures import Future
from typing import TYPE_CHECKING, Any

from coredis.commands.constants import CommandName
from coredis.exceptions import RedisError
from coredis.response.types import MonitorResult
from coredis.typing import AnyStr, Callable, Generator, Generic, Optional, TypeVar

if TYPE_CHECKING:
    import coredis.client
    import coredis.connection

MonitorT = TypeVar("MonitorT", bound="Monitor[Any]")


class Monitor(Generic[AnyStr]):
    """
    Monitor is useful for handling the ``MONITOR`` command to the redis server.

    It can be used as an infinite async iterator::

        async for command in client.monitor():
            print(command.time, command.client_type, command.command, command.args)

    Alternatively, each command can be fetched explicitly::

        monitor = client.monitor()
        command1 = await monitor.get_command()
        command2 = await monitor.get_command()
        monitor.stop()

    """

    def __init__(self, client: coredis.client.Client[AnyStr]):
        self.client: coredis.client.Client[AnyStr] = client
        self.encoding = client.encoding
        self.connection: Optional[coredis.connection.Connection] = None
        self.monitoring = False

    def __aiter__(self) -> Monitor[AnyStr]:
        return self

    async def __anext__(self) -> MonitorResult:
        """
        Infinite iterator that streams back the next command processed by the
        monitored server.
        """
        return await self.get_command()

    def __await__(self: MonitorT) -> Generator[Any, None, MonitorT]:
        return self.__start_monitor().__await__()

    async def get_command(self) -> MonitorResult:
        """
        Wait for the next command issued and return the details
        """
        await self.__start_monitor()
        assert self.connection
        response = await self.connection.fetch_push_message(block=True)
        if isinstance(response, bytes):
            response = response.decode(self.encoding)
        assert isinstance(response, str)
        return MonitorResult.parse_response_string(response)

    async def stop(self) -> None:
        """
        Stop monitoring by issuing a ``RESET`` command
        and release the connection.
        """
        return await self.__stop_monitoring()

    def run_in_thread(
        self,
        response_handler: Callable[[MonitorResult], None],
        loop: Optional[AbstractEventLoop] = None,
    ) -> MonitorThread:
        """
        Runs the monitor in a :class:`MonitorThread` and invokes :paramref:`response_handler`
        for every command received.

        To stop the processing call :meth:`MonitorThread.stop` on the instance
        returned by this method.

        """
        monitor_thread = MonitorThread(
            self, loop or asyncio.get_event_loop(), response_handler
        )
        monitor_thread.start()
        return monitor_thread

    async def __connect(self) -> None:
        if self.connection is None:
            self.connection = await self.client.connection_pool.get_connection()

    async def __start_monitor(self: MonitorT) -> MonitorT:
        if self.monitoring:
            return self
        await self.__connect()
        assert self.connection
        request = await self.connection.create_request(
            CommandName.MONITOR, decode=False
        )
        response = await request
        if not response == b"OK":  # noqa
            raise RedisError(f"Failed to start MONITOR {response!r}")
        self.monitoring = True
        return self

    async def __stop_monitoring(self) -> None:
        if self.connection:
            request = await self.connection.create_request(
                CommandName.RESET, decode=False
            )
            response = await request
            if not response == CommandName.RESET:  # noqa
                raise RedisError("Failed to reset connection")
        self.__reset()

    def __reset(self) -> None:
        if self.connection:
            self.connection.disconnect()
            self.client.connection_pool.release(self.connection)
        self.monitoring = False
        self.connection = None


class MonitorThread(threading.Thread):
    """
    Thread to be used to run monitor
    """

    def __init__(
        self,
        monitor: Monitor[Any],
        loop: asyncio.events.AbstractEventLoop,
        response_handler: Callable[[MonitorResult], None],
    ):
        self._monitor = monitor
        self._loop = loop
        self._response_handler = response_handler
        self._future: Optional[Future[None]] = None
        super().__init__()

    def run(self) -> None:
        self._future = asyncio.run_coroutine_threadsafe(self._run(), self._loop)

    async def _run(self) -> None:
        try:
            async for command in self._monitor:
                self._response_handler(command)
        except CancelledError:
            await self._monitor.stop()

    def stop(self) -> None:
        if self._future:
            self._future.cancel()



from __future__ import annotations

from typing import cast

from coredis._utils import EncodingInsensitiveDict, nativestr
from coredis.response._callbacks import ResponseCallback
from coredis.response._callbacks.server import InfoCallback
from coredis.typing import (
    AnyStr,
    Dict,
    List,
    Optional,
    ResponsePrimitive,
    ResponseType,
    Tuple,
    Union,
    ValueT,
)

SENTINEL_STATE_INT_FIELDS = {
    "can-failover-its-master",
    "config-epoch",
    "down-after-milliseconds",
    "failover-timeout",
    "info-refresh",
    "last-hello-message",
    "last-ok-ping-reply",
    "last-ping-reply",
    "last-ping-sent",
    "master-link-down-time",
    "master-port",
    "num-other-sentinels",
    "num-slaves",
    "o-down-time",
    "pending-commands",
    "parallel-syncs",
    "port",
    "quorum",
    "role-reported-time",
    "s-down-time",
    "slave-priority",
    "slave-repl-offset",
    "voted-leader-epoch",
}


def sentinel_state_typed(
    response: list[str],
) -> Dict[str, Union[str, int, bool]]:
    it = iter(response)
    result: Dict[str, Union[str, int, bool]] = {}

    for key, value in zip(it, it):
        if key in SENTINEL_STATE_INT_FIELDS:
            value = int(value)
        result[key] = value

    return result


def add_flags(
    result: Dict[str, Union[int, str, bool]]
) -> Dict[str, Union[int, str, bool]]:
    flags = set(nativestr(result["flags"]).split(","))
    for name, flag in (
        ("is_master", "master"),
        ("is_slave", "slave"),
        ("is_sdown", "s_down"),
        ("is_odown", "o_down"),
        ("is_sentinel", "sentinel"),
        ("is_disconnected", "disconnected"),
        ("is_master_down", "master_down"),
    ):
        result[name] = flag in flags
    return result


def parse_sentinel_state(
    item: List[ResponsePrimitive],
) -> Dict[str, Union[int, str, bool]]:
    result = sentinel_state_typed([nativestr(k) for k in item])
    result = add_flags(result)
    return result


class PrimaryCallback(
    ResponseCallback[
        ResponseType,
        Dict[ResponsePrimitive, ResponsePrimitive],
        Dict[str, Union[str, int, bool]],
    ]
):
    def transform(
        self, response: ResponseType, **options: Optional[ValueT]
    ) -> Dict[str, Union[str, int, bool]]:
        return parse_sentinel_state(cast(List[ResponsePrimitive], response))

    def transform_3(
        self,
        response: Dict[ResponsePrimitive, ResponsePrimitive],
        **options: Optional[ValueT],
    ) -> Dict[str, Union[str, int, bool]]:
        return add_flags(EncodingInsensitiveDict(response))


class PrimariesCallback(
    ResponseCallback[
        List[ResponseType],
        List[ResponseType],
        Dict[str, Dict[str, Union[str, int, bool]]],
    ]
):
    def transform(
        self,
        response: Union[List[ResponseType], Dict[ResponsePrimitive, ResponsePrimitive]],
        **options: Optional[ValueT],
    ) -> Dict[str, Dict[str, Union[str, int, bool]]]:
        result: Dict[str, Dict[str, Union[str, int, bool]]] = {}

        for item in response:
            state = PrimaryCallback()(item)
            result[str(state["name"])] = state

        return result

    def transform_3(
        self, response: List[ResponseType], **options: Optional[ValueT]
    ) -> Dict[str, Dict[str, Union[str, int, bool]]]:
        states: Dict[str, Dict[str, Union[str, int, bool]]] = {}
        for state in response:
            proxy = add_flags(EncodingInsensitiveDict(state))
            states[nativestr(proxy["name"])] = proxy
        return states


class SentinelsStateCallback(
    ResponseCallback[
        List[ResponseType],
        List[ResponseType],
        Tuple[Dict[str, Union[str, bool, int]], ...],
    ]
):
    def transform(
        self, response: List[ResponseType], **options: Optional[ValueT]
    ) -> Tuple[Dict[str, Union[str, bool, int]], ...]:
        return tuple(
            parse_sentinel_state([nativestr(i) for i in item]) for item in response
        )

    def transform_3(
        self, response: List[ResponseType], **options: Optional[ValueT]
    ) -> Tuple[Dict[str, Union[str, bool, int]], ...]:
        return tuple(add_flags(EncodingInsensitiveDict(state)) for state in response)


class GetPrimaryCallback(
    ResponseCallback[
        List[ResponsePrimitive],
        List[ResponsePrimitive],
        Optional[Tuple[str, int]],
    ]
):
    def transform(
        self, response: List[ResponsePrimitive], **options: Optional[ValueT]
    ) -> Optional[Tuple[str, int]]:
        return nativestr(response[0]), int(response[1]) if response else None


class SentinelInfoCallback(
    ResponseCallback[
        List[ResponseType],
        List[ResponseType],
        Dict[AnyStr, Dict[int, Dict[str, ResponseType]]],
    ]
):
    def transform(
        self, response: List[ResponseType], **options: Optional[ValueT]
    ) -> Dict[AnyStr, Dict[int, Dict[str, ResponseType]]]:
        return {response[0]: {r[0]: InfoCallback()(r[1]) for r in response[1]}}



from __future__ import annotations

from coredis.modules.response.types import AutocompleteSuggestion
from coredis.response._callbacks import ResponseCallback
from coredis.typing import AnyStr, List, Optional, ResponseType, Tuple, Union, ValueT


class AutocompleteCallback(
    ResponseCallback[
        List[ResponseType],
        List[ResponseType],
        Union[Tuple[AutocompleteSuggestion[AnyStr], ...], Tuple[()]],
    ]
):
    def transform(
        self, response: List[ResponseType], **options: Optional[ValueT]
    ) -> Union[Tuple[AutocompleteSuggestion[AnyStr], ...], Tuple[()]]:
        if not response:
            return ()
        step = 1
        results = []
        score_idx = payload_idx = 0
        if options.get("withscores"):
            score_idx = 1
            step += 1
        if options.get("withpayloads"):
            payload_idx = score_idx + 1
            step += 1

        for k in range(0, len(response), step):
            section = response[k : k + step]
            score = section[score_idx] if score_idx else None
            results.append(
                AutocompleteSuggestion(
                    section[0],
                    float(score) if score else None,
                    section[payload_idx] if payload_idx else None,
                )
            )

        return tuple(results)



from __future__ import annotations

from deprecated.sphinx import versionadded

from ..commands._wrappers import CacheConfig
from ..commands.constants import CommandFlag, CommandGroup, CommandName
from ..response._callbacks import BoolCallback, IntCallback
from ..tokens import PrefixToken, PureToken
from ..typing import AnyStr, CommandArgList, KeyT, Optional, StringT, Tuple, Union
from .base import ModuleGroup, module_command
from .response._callbacks.autocomplete import AutocompleteCallback
from .response.types import AutocompleteSuggestion
from .search import RediSearch


@versionadded(version="4.12")
class Autocomplete(ModuleGroup[AnyStr]):
    MODULE = RediSearch
    COMMAND_GROUP = CommandGroup.SUGGESTION

    @module_command(
        CommandName.FT_SUGADD,
        module=MODULE,
        version_introduced="1.0.0",
        group=COMMAND_GROUP,
    )
    async def sugadd(
        self,
        key: KeyT,
        string: StringT,
        score: Union[int, float],
        increment_score: Optional[bool] = None,
        payload: Optional[StringT] = None,
    ) -> int:
        """
        Adds a suggestion string to an auto-complete suggestion dictionary

        :param key: The suggestion dictionary key.
        :param string: The suggestion string to index.
        :param score: The floating point number of the suggestion string's weight.
        :param increment_score: Increments the existing entry of the suggestion by
         the given score, instead of replacing the score.
        :param payload: Saves an extra payload with the suggestion, that can be
         fetched when calling :meth:`sugget` by using :paramref:`sugget.withpayloads`
        """
        pieces: CommandArgList = [key, string, score]
        if increment_score:
            pieces.append(PureToken.INCREMENT)
        if payload:
            pieces.extend([PrefixToken.PAYLOAD, payload])

        return await self.execute_module_command(
            CommandName.FT_SUGADD, *pieces, callback=IntCallback()
        )

    @module_command(
        CommandName.FT_SUGGET,
        module=MODULE,
        version_introduced="1.0.0",
        group=COMMAND_GROUP,
        cache_config=CacheConfig(lambda *a, **_: a[0]),
        flags={CommandFlag.READONLY},
    )
    async def sugget(
        self,
        key: KeyT,
        prefix: StringT,
        *,
        fuzzy: Optional[bool] = None,
        withscores: Optional[bool] = None,
        withpayloads: Optional[bool] = None,
        max_suggestions: Optional[int] = None,
    ) -> Union[Tuple[AutocompleteSuggestion[AnyStr], ...], Tuple[()]]:
        """
        Gets completion suggestions for a prefix

        :param key: The suggestion dictionary key.
        :param prefix: The prefix to complete on.
        :param fuzzy: If ``True``, performs a fuzzy prefix search, including prefixes at
         Levenshtein distance of 1 from the prefix sent.
        :param withscores: If ``True``, also returns the score of each suggestion.
        :param withpayloads: If True, returns optional payloads saved along with the suggestions.
        :param max_suggestions: Limits the results to a maximum of ``max_suggestions``
        """
        pieces: CommandArgList = [key, prefix]
        if fuzzy:
            pieces.append(PureToken.FUZZY)
        if withscores:
            pieces.append(PureToken.WITHSCORES)
        if withpayloads:
            pieces.append(PureToken.WITHPAYLOADS)
        if max_suggestions is not None:
            pieces.append(PureToken.MAX)
            pieces.append(max_suggestions)

        return await self.execute_module_command(
            CommandName.FT_SUGGET,
            *pieces,
            callback=AutocompleteCallback[AnyStr](),
            withscores=withscores,
            withpayloads=withpayloads,
        )

    @module_command(
        CommandName.FT_SUGDEL,
        module=MODULE,
        version_introduced="1.0.0",
        group=COMMAND_GROUP,
    )
    async def sugdel(self, key: KeyT, string: StringT) -> bool:
        """
        Deletes a string from a suggestion index

        :param key: The suggestion dictionary key.
        :param string: The suggestion string to index.

        """
        pieces: CommandArgList = [key, string]

        return await self.execute_module_command(
            CommandName.FT_SUGDEL, *pieces, callback=BoolCallback()
        )

    @module_command(
        CommandName.FT_SUGLEN,
        module=MODULE,
        version_introduced="1.0.0",
        group=COMMAND_GROUP,
    )
    async def suglen(self, key: KeyT) -> int:
        """
        Gets the size of an auto-complete suggestion dictionary

        :param key: The key of the suggestion dictionary.
        """
        pieces: CommandArgList = [key]

        return await self.execute_module_command(
            CommandName.FT_SUGLEN, *pieces, callback=IntCallback()
        )



from __future__ import annotations

from typing import cast

from coredis.typing import (
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    OrderedDict,
    StringT,
    Tuple,
    TypeVar,
    Union,
)

T_co = TypeVar("T_co")


def flat_pairs_to_dict(
    response: Union[Tuple[T_co, ...], List[T_co]],
    value_transform: Optional[Callable[..., T_co]] = None,
) -> Dict[T_co, T_co]:
    """Creates a dict given a flat list of key/value pairs"""
    if isinstance(response, Dict):
        return response
    it = iter(response)
    if value_transform:
        return dict(zip(it, map(value_transform, it)))
    else:
        return dict(zip(it, it))


def flat_pairs_to_ordered_dict(response: Iterable[T_co]) -> OrderedDict[StringT, T_co]:
    """Creates a dict given a flat list of key/value pairs"""
    it = iter(response)
    return cast(OrderedDict[StringT, T_co], OrderedDict(zip(it, it)))



from __future__ import annotations

from typing import cast

from coredis._utils import EncodingInsensitiveDict
from coredis.response._callbacks import ResponseCallback
from coredis.response._utils import flat_pairs_to_dict
from coredis.response.types import LibraryDefinition
from coredis.typing import (
    AnyStr,
    Dict,
    List,
    Mapping,
    Optional,
    ResponsePrimitive,
    ResponseType,
    Union,
    ValueT,
)


class FunctionListCallback(
    ResponseCallback[
        List[ResponseType], List[ResponseType], Mapping[str, LibraryDefinition]
    ]
):
    def transform(
        self, response: List[ResponseType], **options: Optional[ValueT]
    ) -> Mapping[str, LibraryDefinition]:
        libraries = [
            EncodingInsensitiveDict(flat_pairs_to_dict(cast(List[ValueT], library)))
            for library in response
        ]
        transformed = EncodingInsensitiveDict()
        for library in libraries:
            lib_name = library["library_name"]
            functions = EncodingInsensitiveDict({})
            for function in library.get("functions", []):
                function_definition = EncodingInsensitiveDict(
                    flat_pairs_to_dict(function)
                )
                functions[function_definition["name"]] = function_definition
                functions[function_definition["name"]]["flags"] = set(
                    function_definition["flags"]
                )
            library["functions"] = functions
            transformed[lib_name] = EncodingInsensitiveDict(
                LibraryDefinition(
                    name=library["name"],
                    engine=library["engine"],
                    description=library["description"],
                    functions=library["functions"],
                    library_code=library["library_code"],
                )
            )
        return transformed


class FunctionStatsCallback(
    ResponseCallback[
        List[ResponseType],
        Dict[AnyStr, Union[AnyStr, Dict[AnyStr, Dict[AnyStr, ResponsePrimitive]]]],
        Dict[AnyStr, Union[AnyStr, Dict[AnyStr, Dict[AnyStr, ResponsePrimitive]]]],
    ]
):
    def transform(
        self,
        response: List[ResponseType],
        **options: Optional[ValueT],
    ) -> Dict[AnyStr, Union[AnyStr, Dict[AnyStr, Dict[AnyStr, ResponsePrimitive]]]]:
        transformed = flat_pairs_to_dict(response)
        key = cast(AnyStr, b"engines" if b"engines" in transformed else "engines")
        engines = flat_pairs_to_dict(cast(List[AnyStr], transformed.pop(key)))
        engines_transformed = {}
        for engine, stats in engines.items():
            engines_transformed[engine] = flat_pairs_to_dict(cast(List[AnyStr], stats))
        transformed[key] = engines_transformed  # type: ignore
        return cast(
            Dict[AnyStr, Union[AnyStr, Dict[AnyStr, Dict[AnyStr, ResponsePrimitive]]]],
            transformed,
        )

    def transform_3(
        self,
        response: Dict[
            AnyStr, Union[AnyStr, Dict[AnyStr, Dict[AnyStr, ResponsePrimitive]]]
        ],
        **options: Optional[ValueT],
    ) -> Dict[AnyStr, Union[AnyStr, Dict[AnyStr, Dict[AnyStr, ResponsePrimitive]]]]:
        return response



from __future__ import annotations

from deprecated.sphinx import versionadded

from .._json import json
from ..commands._wrappers import CacheConfig
from ..commands.constants import CommandFlag, CommandGroup, CommandName
from ..response._callbacks import (
    IntCallback,
    NoopCallback,
    OneOrManyCallback,
    SimpleStringCallback,
)
from ..tokens import PureToken
from ..typing import (
    AnyStr,
    CommandArgList,
    KeyT,
    List,
    Literal,
    Optional,
    Parameters,
    ResponseType,
    StringT,
    Tuple,
    Union,
    ValueT,
)
from .base import Module, ModuleGroup, module_command
from .response._callbacks.json import JsonCallback
from .response.types import JsonType


class RedisJSON(Module[AnyStr]):
    NAME = "ReJSON"
    FULL_NAME = "RedisJSON"
    DESCRIPTION = """RedisJSON is a Redis module that implements a JSON data type 
and a set of commands to operate on it."""
    DOCUMENTATION_URL = "https://redis.io/docs/stack/json/"


@versionadded(version="4.12")
class Json(ModuleGroup[AnyStr]):
    MODULE = RedisJSON
    COMMAND_GROUP = CommandGroup.JSON

    @module_command(
        CommandName.JSON_DEL,
        group=COMMAND_GROUP,
        version_introduced="1.0.0",
        module=MODULE,
    )
    async def delete(self, key: KeyT, path: Optional[StringT] = None) -> int:
        """
        Delete a value from a JSON document.

        :param key: The key of the JSON document.
        :param path: The JSONPath to specify.
        :return: The number of paths deleted
        """
        pieces: CommandArgList = [key]
        if path:
            pieces.append(path)

        return await self.execute_module_command(
            CommandName.JSON_DEL, *pieces, callback=IntCallback()
        )

    @module_command(
        CommandName.JSON_GET,
        group=COMMAND_GROUP,
        version_introduced="1.0.0",
        module=MODULE,
        cache_config=CacheConfig(lambda *a, **_: a[0]),
        flags={CommandFlag.READONLY},
    )
    async def get(
        self,
        key: KeyT,
        *paths: StringT,
    ) -> JsonType:
        """
        Gets the value at one or more paths

        :param key: The key of the JSON document.
        :param paths: JSONPath(s) to get values from.
        :return: The value at :paramref:`path`
        """
        pieces: CommandArgList = [key]
        if paths:
            pieces.extend(paths)

        return await self.execute_module_command(
            CommandName.JSON_GET, *pieces, callback=JsonCallback()
        )

    @module_command(
        CommandName.JSON_FORGET,
        group=COMMAND_GROUP,
        version_introduced="1.0.0",
        module=MODULE,
    )
    async def forget(self, key: KeyT, path: Optional[ValueT] = None) -> int:
        """
        Deletes an element from a path from a json object

        :param key: The key of the JSON document.
        :param path: The path(s) to delete from the JSON object.

        :return: The number of deleted elements.
        """
        pieces: CommandArgList = [key]
        if path:
            pieces.append(path)
        return await self.execute_module_command(
            CommandName.JSON_FORGET, *pieces, callback=IntCallback()
        )

    @module_command(
        CommandName.JSON_TOGGLE,
        group=COMMAND_GROUP,
        version_introduced="2.0.0",
        module=MODULE,
    )
    async def toggle(self, key: KeyT, path: ValueT) -> JsonType:
        """
        Toggles a boolean value

        :param key: Redis key to modify.
        :param path: JSONPath to specify.
        :return: A list of integer replies for each path, the new value
         (`0` if `false` or `1` if `true`), or ``None`` for JSON values matching
         the path that are not Boolean.
        """
        pieces: CommandArgList = [key, path]
        return await self.execute_module_command(
            CommandName.JSON_TOGGLE,
            *pieces,
            callback=JsonCallback(),
        )

    @module_command(
        CommandName.JSON_CLEAR,
        group=COMMAND_GROUP,
        version_introduced="2.0.0",
        module=MODULE,
    )
    async def clear(self, key: KeyT, path: Optional[ValueT] = None) -> int:
        """
        Clears all values from an array or an object and sets numeric values to `0`

        :param key: The key to parse.
        :param path: The JSONPath to specify.
        :return: The number of values cleared.
        """
        pieces: CommandArgList = [key]
        if path:
            pieces.append(path)

        return await self.execute_module_command(
            CommandName.JSON_CLEAR, *pieces, callback=IntCallback()
        )

    @module_command(
        CommandName.JSON_SET,
        group=COMMAND_GROUP,
        version_introduced="1.0.0",
        module=MODULE,
    )
    async def set(
        self,
        key: KeyT,
        path: ValueT,
        value: JsonType,
        condition: Optional[Literal[PureToken.NX, PureToken.XX]] = None,
    ) -> bool:
        """
        Sets or updates the JSON value at a path

        :param key: The key to parse.
        :param path: JSONPath to specify. For new Redis keys the ``path`` must be the root.
         For existing keys, when the entire `path` exists, the value that it contains is replaced
         with :paramref:`value`. For existing keys, when the ``path`` exists, except for the
         last element, a new child is added with :paramref:`value`.
         Adds a key (with its respective value) to a JSON Object (in a json data type key)
         only if it is the last child in the ``path``, or it is the parent of a new child being
         added in the ``path``. Optional argument :paramref:`condition` modifies this behavior
         for both new json data type keys as well as the JSON Object keys in them.
        :param value: Value to set at the specified :paramref:`path`.
        :param condition: Optional argument to modify the behavior of adding a key to a JSON Object.
         If ``NX``, the key is set only if it does not already exist. If ``XX``, the key is set only
         if it already exists.
        :return: `True` if the value was set successfully, `False` otherwise.
        """
        pieces: CommandArgList = [key, path, json.dumps(value)]
        if condition:
            pieces.append(condition)
        return await self.execute_module_command(
            CommandName.JSON_SET, *pieces, callback=SimpleStringCallback()
        )

    @module_command(
        CommandName.JSON_MGET,
        group=COMMAND_GROUP,
        version_introduced="1.0.0",
        module=MODULE,
        flags={CommandFlag.READONLY},
    )
    async def mget(self, keys: Parameters[KeyT], path: StringT) -> JsonType:
        """
        Returns the values at a path from one or more keys

        :param keys: one or more keys to retrieve values from.
        :param path: JSONPath to specify.
        :return: The values at :paramref:`path` for each of the keys in :paramref:`keys`.
        """
        pieces: CommandArgList = [*keys, path]
        return await self.execute_module_command(
            CommandName.JSON_MGET,
            *pieces,
            callback=JsonCallback(),
            keys=keys,  # type: ignore
        )

    @module_command(
        CommandName.JSON_MSET,
        group=COMMAND_GROUP,
        version_introduced="2.6.0",
        module=MODULE,
    )
    async def mset(self, triplets: Parameters[Tuple[KeyT, StringT, JsonType]]) -> bool:
        """
        Sets or updates the JSON value of one or more keys

        :param triplets: Collection of triplets containing (``key``, ``path``, ``value``)
         to set.

        :return: `True` if all the values were set successfully
        """
        pieces: CommandArgList = []
        for key, path, value in triplets:
            pieces.extend([key, path, json.dumps(value)])

        return await self.execute_module_command(
            CommandName.JSON_MSET, *pieces, callback=SimpleStringCallback()
        )

    @module_command(
        CommandName.JSON_MERGE,
        group=COMMAND_GROUP,
        version_introduced="2.6.0",
        module=MODULE,
    )
    async def merge(self, key: KeyT, path: StringT, value: JsonType) -> bool:
        """
        Merge a JSON object into an existing Redis key at a specified path.

        :param key: The Redis key to merge the JSON object into.
        :param path: The JSONPath within the Redis key to merge the JSON object into.
        :param value: The JSON object to merge into the Redis key.
        :return: True if the merge was successful, False otherwise.
        """
        pieces: CommandArgList = [key, path, json.dumps(value)]

        return await self.execute_module_command(
            CommandName.JSON_MERGE, *pieces, callback=SimpleStringCallback()
        )

    @module_command(
        CommandName.JSON_NUMINCRBY,
        group=COMMAND_GROUP,
        version_introduced="1.0.0",
        module=MODULE,
    )
    async def numincrby(
        self, key: KeyT, path: ValueT, value: Union[int, float]
    ) -> JsonType:
        """
        Increments the numeric value at path by a value

        :param key: The key to modify.
        :param path: The JSONPath to specify.
        :param value: The number value to increment.
        """
        pieces: CommandArgList = [key, path, value]

        return await self.execute_module_command(
            CommandName.JSON_NUMINCRBY, *pieces, callback=JsonCallback()
        )

    @module_command(
        CommandName.JSON_NUMMULTBY,
        group=COMMAND_GROUP,
        version_introduced="1.0.0",
        module=MODULE,
    )
    async def nummultby(
        self, key: KeyT, path: ValueT, value: Union[int, float]
    ) -> JsonType:
        """
        Multiplies the numeric value at path by a value

        :param key: Key to modify.
        :param path: JSONPath to specify.
        :param value: Number value to multiply.
        """
        pieces: CommandArgList = [key, path, value]

        return await self.execute_module_command(
            CommandName.JSON_NUMMULTBY, *pieces, callback=JsonCallback()
        )

    @module_command(
        CommandName.JSON_STRAPPEND,
        group=COMMAND_GROUP,
        version_introduced="1.0.0",
        module=MODULE,
    )
    async def strappend(
        self,
        key: KeyT,
        value: Optional[Union[str, bytes, int, float]],
        path: Optional[KeyT] = None,
    ) -> Optional[Union[int, List[Optional[int]]]]:
        """
        Appends a string to a JSON string value at path

        :param key: The key to modify
        :param value: The value to append to the string(s) at the specified :paramref:`path`.
        :param path: The JSONPath to specify the location of the string(s) to modify.
        :return: A list of integer replies for each path, the string's new length,
         or ``None`` if the matching JSON value is not a string.
        """
        pieces: CommandArgList = [key]
        if path is not None:
            pieces.append(path)
        pieces.append(json.dumps(value))
        return await self.execute_module_command(
            CommandName.JSON_STRAPPEND, *pieces, callback=OneOrManyCallback[int]()
        )

    @module_command(
        CommandName.JSON_STRLEN,
        group=COMMAND_GROUP,
        version_introduced="1.0.0",
        module=MODULE,
        flags={CommandFlag.READONLY},
        cache_config=CacheConfig(lambda *a, **_: a[0]),
    )
    async def strlen(
        self, key: KeyT, path: Optional[KeyT] = None
    ) -> Optional[Union[int, List[Optional[int]]]]:
        """
        Returns the length of the JSON String at path in key

        :param key: The key of the JSON document.
        :param path: JSONPath to specify.
        :return: An array of integer replies for each path, the array's length,
         or ``None``, if the matching JSON value is not a string.


        """
        pieces: CommandArgList = [key]
        if path is not None:
            pieces.append(path)

        return await self.execute_module_command(
            CommandName.JSON_STRLEN, *pieces, callback=OneOrManyCallback[int]()
        )

    @module_command(
        CommandName.JSON_ARRAPPEND,
        group=COMMAND_GROUP,
        version_introduced="1.0.0",
        module=MODULE,
    )
    async def arrappend(
        self,
        key: KeyT,
        values: Parameters[JsonType],
        path: Optional[KeyT] = None,
    ) -> Optional[Union[int, List[Optional[int]]]]:
        """
        Append one or more json values into the array at path after the last element in it.

        :param key: The key to modify.
        :param values: One or more values to append to one or more arrays.
        :param path: JSONPath to specify.
        :return: An array of integer replies for each path, the array's new size,
         or `None` if the matching JSON value is not an array.

        """
        pieces: CommandArgList = [key]
        if path:
            pieces.append(path)
        pieces.extend([json.dumps(value) for value in values])
        return await self.execute_module_command(
            CommandName.JSON_ARRAPPEND,
            *pieces,
            callback=OneOrManyCallback[int](),
        )

    @module_command(
        CommandName.JSON_ARRINDEX,
        group=COMMAND_GROUP,
        version_introduced="1.0.0",
        module=MODULE,
        flags={CommandFlag.READONLY},
        cache_config=CacheConfig(lambda *a, **_: a[0]),
    )
    async def arrindex(
        self,
        key: KeyT,
        path: ValueT,
        value: Union[str, bytes, int, float],
        start: Optional[int] = None,
        stop: Optional[int] = None,
    ) -> Optional[Union[int, List[Optional[int]]]]:
        """
        Returns the index of the first occurrence of a JSON scalar value in the array at path

        :param key: The key to parse.
        :param path: The JSONPath to specify.
        :param value: The value to find its index in one or more arrays.
        :param start: Inclusive start value to specify in a slice of the array to search.
        :param stop: Exclusive stop value to specify in a slice of the array to search,
         including the last element. Negative values are interpreted as starting from the end.
        :return: The index of the first occurrence of the value in the array,
         or a list of indices if the value is found in multiple arrays.
        """
        pieces: CommandArgList = [key, path, json.dumps(value)]
        if start is not None:
            pieces.append(start)
        if stop is not None:
            pieces.append(stop)

        return await self.execute_module_command(
            CommandName.JSON_ARRINDEX, *pieces, callback=OneOrManyCallback[int]()
        )

    @module_command(
        CommandName.JSON_ARRINSERT,
        group=COMMAND_GROUP,
        version_introduced="1.0.0",
        module=MODULE,
    )
    async def arrinsert(
        self,
        key: KeyT,
        path: ValueT,
        index: int,
        values: Parameters[JsonType],
    ) -> Optional[Union[int, List[Optional[int]]]]:
        """
        Inserts the JSON scalar(s) value at the specified index in the array at path

        :param key: Key to modify.
        :param path: JSONPath to specify.
        :param index: Position in the array where you want to insert a value.
         The index must be in the array's range. Inserting at `index` 0 prepends to the array.
         Negative index values start from the end of the array.
        :param values: One or more values to insert in one or more arrays.
        :returns: The length of the array after the insert operation or a list of lengths of
         the arrays after the insert operation if the path matches multiple arrays
        """
        pieces: CommandArgList = [key, path, index]
        pieces.extend([json.dumps(value) for value in values])

        return await self.execute_module_command(
            CommandName.JSON_ARRINSERT, *pieces, callback=OneOrManyCallback[int]()
        )

    @module_command(
        CommandName.JSON_ARRLEN,
        group=COMMAND_GROUP,
        version_introduced="1.0.0",
        module=MODULE,
        flags={CommandFlag.READONLY},
        cache_config=CacheConfig(lambda *a, **_: a[0]),
    )
    async def arrlen(
        self, key: KeyT, path: Optional[KeyT] = None
    ) -> Optional[Union[int, List[Optional[int]]]]:
        """
        Returns the length of the array at path

        :param key: The key to parse.
        :param path: The JSONPath to specify.

        :return: An integer if the matching value is an array, or a list of integers if
         multiple matching values are arrays. Returns ``None`` if the :paramref:`key` or
         :paramref:`path` do not exist.
        """
        pieces: CommandArgList = [key]
        if path:
            pieces.append(path)

        return await self.execute_module_command(
            CommandName.JSON_ARRLEN, *pieces, callback=OneOrManyCallback[int]()
        )

    @module_command(
        CommandName.JSON_ARRPOP,
        group=COMMAND_GROUP,
        version_introduced="1.0.0",
        module=MODULE,
    )
    async def arrpop(
        self, key: KeyT, path: Optional[KeyT] = None, index: Optional[int] = None
    ) -> JsonType:
        """
        Removes and returns the element at the specified index in the array at path

        :param key: Key to modify.
        :param path: JSONPath to specify.
        :param index: Position in the array to start popping from. Out-of-range indexes
         round to their respective array ends.
        :return: The popped value, or ``None`` if the matching JSON value is not an array.
        """
        pieces: CommandArgList = [key]
        if path:
            pieces.append(path)
        if index is not None:
            pieces.append(index)

        return await self.execute_module_command(
            CommandName.JSON_ARRPOP, *pieces, callback=JsonCallback()
        )

    @module_command(
        CommandName.JSON_ARRTRIM,
        group=COMMAND_GROUP,
        version_introduced="1.0.0",
        module=MODULE,
    )
    async def arrtrim(
        self, key: KeyT, path: ValueT, start: int, stop: int
    ) -> Optional[Union[int, List[Optional[int]]]]:
        """
        Trims the array at path to contain only the specified inclusive range of indices
        from start to stop

        :param key: Key to modify.
        :param path: The JSONPath to specify.
        :param start: The index of the first element to keep (previous elements are trimmed).
        :param stop: The index of the last element to keep (following elements are trimmed),
         including the last element. Negative values are interpreted as starting from the end.
        :return: The number of elements removed or a list if multiple matching values are arrays.
        """
        pieces: CommandArgList = [key, path, start, stop]

        return await self.execute_module_command(
            CommandName.JSON_ARRTRIM, *pieces, callback=OneOrManyCallback[int]()
        )

    @module_command(
        CommandName.JSON_OBJKEYS,
        group=COMMAND_GROUP,
        version_introduced="1.0.0",
        module=MODULE,
    )
    async def objkeys(self, key: KeyT, path: Optional[StringT] = None) -> ResponseType:
        """
        Returns the JSON keys of the object at path

        :param key: The key of the JSON document.
        :param path: JSONPath to specify.

        :return: A list of the key names in the object or a list of lists if multiple objects
         match the :paramref:`path`, or `None` if the matching value is not an object.

        """
        pieces: CommandArgList = [key]
        if path:
            pieces.append(path)

        return await self.execute_module_command(
            CommandName.JSON_OBJKEYS, *pieces, callback=NoopCallback[ResponseType]()
        )

    @module_command(
        CommandName.JSON_OBJLEN,
        group=COMMAND_GROUP,
        version_introduced="1.0.0",
        module=MODULE,
    )
    async def objlen(
        self, key: KeyT, path: Optional[KeyT] = None
    ) -> Optional[Union[int, List[Optional[int]]]]:
        """
        Returns the number of keys of the object at path

        :param key: The key of the JSON document.
        :param path: JSONPath to specify.
        :return: An integer if the path matches exactly one object or a list of integer
         replies for each path specified as the number of keys in the object or ``None``,
         if the matching JSON value is not an object.
        """
        pieces: CommandArgList = [key]
        if path:
            pieces.append(path)

        return await self.execute_module_command(
            CommandName.JSON_OBJLEN, *pieces, callback=OneOrManyCallback[int]()
        )

    @module_command(
        CommandName.JSON_TYPE,
        group=COMMAND_GROUP,
        version_introduced="1.0.0",
        module=MODULE,
        flags={CommandFlag.READONLY},
        cache_config=CacheConfig(lambda *a, **_: a[0]),
    )
    async def type(
        self, key: KeyT, path: Optional[KeyT] = None
    ) -> Optional[Union[AnyStr, List[Optional[AnyStr]]]]:
        """
        Returns the type of the JSON value at path

        :param key: The key to parse.
        :param path: The JSONPath to specify.
        """
        pieces: CommandArgList = [key]
        if path is not None:
            pieces.append(path)

        return await self.execute_module_command(
            CommandName.JSON_TYPE, *pieces, callback=OneOrManyCallback[AnyStr]()
        )

    @module_command(
        CommandName.JSON_RESP,
        group=COMMAND_GROUP,
        version_introduced="1.0.0",
        module=MODULE,
        flags={CommandFlag.READONLY},
    )
    async def resp(self, key: KeyT, path: Optional[KeyT] = None) -> ResponseType:
        """
        Returns the JSON value at path in Redis Serialization Protocol (RESP)

        :param key: The key to parse.
        :param path: The JSONPath to specify.
        """
        pieces: CommandArgList = [key]
        if path:
            pieces.append(path)

        return await self.execute_module_command(
            CommandName.JSON_RESP, *pieces, callback=NoopCallback[ResponseType]()
        )

    @module_command(
        CommandName.JSON_DEBUG_MEMORY,
        group=COMMAND_GROUP,
        version_introduced="1.0.0",
        module=MODULE,
        flags={CommandFlag.READONLY},
    )
    async def debug_memory(
        self, key: KeyT, path: Optional[KeyT] = None
    ) -> Optional[Union[int, List[Optional[int]]]]:
        """
        Reports the size in bytes of a key
        """
        pieces: CommandArgList = [key]
        if path:
            pieces.append(path)

        return await self.execute_module_command(
            CommandName.JSON_DEBUG_MEMORY, *pieces, callback=OneOrManyCallback[int]()
        )



"""
coredis.commands.constants
--------------------------
Constants relating to redis command names and groups
"""

from __future__ import annotations

import enum

from coredis._utils import CaseAndEncodingInsensitiveEnum


class CommandName(CaseAndEncodingInsensitiveEnum):
    """
    Enum for listing all redis commands
    """

    #: Commands for server
    BGREWRITEAOF = b"BGREWRITEAOF"  # Since redis: 1.0.0
    BGSAVE = b"BGSAVE"  # Since redis: 1.0.0
    DBSIZE = b"DBSIZE"  # Since redis: 1.0.0
    DEBUG = b"DEBUG"  # Since redis: 1.0.0
    FLUSHALL = b"FLUSHALL"  # Since redis: 1.0.0
    FLUSHDB = b"FLUSHDB"  # Since redis: 1.0.0
    INFO = b"INFO"  # Since redis: 1.0.0
    LASTSAVE = b"LASTSAVE"  # Since redis: 1.0.0
    MONITOR = b"MONITOR"  # Since redis: 1.0.0
    SAVE = b"SAVE"  # Since redis: 1.0.0
    SHUTDOWN = b"SHUTDOWN"  # Since redis: 1.0.0
    SYNC = b"SYNC"  # Since redis: 1.0.0
    CONFIG = b"CONFIG"  # Since redis: 2.0.0
    CONFIG_GET = b"CONFIG GET"  # Since redis: 2.0.0
    CONFIG_RESETSTAT = b"CONFIG RESETSTAT"  # Since redis: 2.0.0
    CONFIG_SET = b"CONFIG SET"  # Since redis: 2.0.0
    SLOWLOG = b"SLOWLOG"  # Since redis: 2.2.12
    SLOWLOG_GET = b"SLOWLOG GET"  # Since redis: 2.2.12
    SLOWLOG_LEN = b"SLOWLOG LEN"  # Since redis: 2.2.12
    SLOWLOG_RESET = b"SLOWLOG RESET"  # Since redis: 2.2.12
    TIME = b"TIME"  # Since redis: 2.6.0
    CONFIG_REWRITE = b"CONFIG REWRITE"  # Since redis: 2.8.0
    PSYNC = b"PSYNC"  # Since redis: 2.8.0
    ROLE = b"ROLE"  # Since redis: 2.8.12
    COMMAND = b"COMMAND"  # Since redis: 2.8.13
    COMMAND_COUNT = b"COMMAND COUNT"  # Since redis: 2.8.13
    COMMAND_GETKEYS = b"COMMAND GETKEYS"  # Since redis: 2.8.13
    COMMAND_INFO = b"COMMAND INFO"  # Since redis: 2.8.13
    LATENCY = b"LATENCY"  # Since redis: 2.8.13
    LATENCY_DOCTOR = b"LATENCY DOCTOR"  # Since redis: 2.8.13
    LATENCY_GRAPH = b"LATENCY GRAPH"  # Since redis: 2.8.13
    LATENCY_HELP = b"LATENCY HELP"  # Since redis: 2.8.13
    LATENCY_HISTORY = b"LATENCY HISTORY"  # Since redis: 2.8.13
    LATENCY_LATEST = b"LATENCY LATEST"  # Since redis: 2.8.13
    LATENCY_RESET = b"LATENCY RESET"  # Since redis: 2.8.13
    REPLCONF = b"REPLCONF"  # Since redis: 3.0.0
    RESTORE_ASKING = b"RESTORE-ASKING"  # Since redis: 3.0.0
    MEMORY = b"MEMORY"  # Since redis: 4.0.0
    MEMORY_DOCTOR = b"MEMORY DOCTOR"  # Since redis: 4.0.0
    MEMORY_HELP = b"MEMORY HELP"  # Since redis: 4.0.0
    MEMORY_MALLOC_STATS = b"MEMORY MALLOC-STATS"  # Since redis: 4.0.0
    MEMORY_PURGE = b"MEMORY PURGE"  # Since redis: 4.0.0
    MEMORY_STATS = b"MEMORY STATS"  # Since redis: 4.0.0
    MEMORY_USAGE = b"MEMORY USAGE"  # Since redis: 4.0.0
    MODULE = b"MODULE"  # Since redis: 4.0.0
    MODULE_LIST = b"MODULE LIST"  # Since redis: 4.0.0
    MODULE_LOAD = b"MODULE LOAD"  # Since redis: 4.0.0
    MODULE_UNLOAD = b"MODULE UNLOAD"  # Since redis: 4.0.0
    SWAPDB = b"SWAPDB"  # Since redis: 4.0.0
    COMMAND_HELP = b"COMMAND HELP"  # Since redis: 5.0.0
    CONFIG_HELP = b"CONFIG HELP"  # Since redis: 5.0.0
    LOLWUT = b"LOLWUT"  # Since redis: 5.0.0
    MODULE_HELP = b"MODULE HELP"  # Since redis: 5.0.0
    REPLICAOF = b"REPLICAOF"  # Since redis: 5.0.0
    ACL = b"ACL"  # Since redis: 6.0.0
    ACL_CAT = b"ACL CAT"  # Since redis: 6.0.0
    ACL_DELUSER = b"ACL DELUSER"  # Since redis: 6.0.0
    ACL_GENPASS = b"ACL GENPASS"  # Since redis: 6.0.0
    ACL_GETUSER = b"ACL GETUSER"  # Since redis: 6.0.0
    ACL_HELP = b"ACL HELP"  # Since redis: 6.0.0
    ACL_LIST = b"ACL LIST"  # Since redis: 6.0.0
    ACL_LOAD = b"ACL LOAD"  # Since redis: 6.0.0
    ACL_LOG = b"ACL LOG"  # Since redis: 6.0.0
    ACL_SAVE = b"ACL SAVE"  # Since redis: 6.0.0
    ACL_SETUSER = b"ACL SETUSER"  # Since redis: 6.0.0
    ACL_USERS = b"ACL USERS"  # Since redis: 6.0.0
    ACL_WHOAMI = b"ACL WHOAMI"  # Since redis: 6.0.0
    FAILOVER = b"FAILOVER"  # Since redis: 6.2.0
    SLOWLOG_HELP = b"SLOWLOG HELP"  # Since redis: 6.2.0
    ACL_DRYRUN = b"ACL DRYRUN"  # Since redis: 7.0.0
    COMMAND_DOCS = b"COMMAND DOCS"  # Since redis: 7.0.0
    COMMAND_GETKEYSANDFLAGS = b"COMMAND GETKEYSANDFLAGS"  # Since redis: 7.0.0
    COMMAND_LIST = b"COMMAND LIST"  # Since redis: 7.0.0
    LATENCY_HISTOGRAM = b"LATENCY HISTOGRAM"  # Since redis: 7.0.0
    MODULE_LOADEX = b"MODULE LOADEX"  # Since redis: 7.0.0
    SLAVEOF = b"SLAVEOF"  # Deprecated in redis: 5.0.0

    #: Commands for string
    DECR = b"DECR"  # Since redis: 1.0.0
    DECRBY = b"DECRBY"  # Since redis: 1.0.0
    GET = b"GET"  # Since redis: 1.0.0
    INCR = b"INCR"  # Since redis: 1.0.0
    INCRBY = b"INCRBY"  # Since redis: 1.0.0
    MGET = b"MGET"  # Since redis: 1.0.0
    SET = b"SET"  # Since redis: 1.0.0
    MSET = b"MSET"  # Since redis: 1.0.1
    MSETNX = b"MSETNX"  # Since redis: 1.0.1
    APPEND = b"APPEND"  # Since redis: 2.0.0
    SETRANGE = b"SETRANGE"  # Since redis: 2.2.0
    STRLEN = b"STRLEN"  # Since redis: 2.2.0
    GETRANGE = b"GETRANGE"  # Since redis: 2.4.0
    INCRBYFLOAT = b"INCRBYFLOAT"  # Since redis: 2.6.0
    GETDEL = b"GETDEL"  # Since redis: 6.2.0
    GETEX = b"GETEX"  # Since redis: 6.2.0
    LCS = b"LCS"  # Since redis: 7.0.0
    GETSET = b"GETSET"  # Deprecated in redis: 6.2.0
    SETNX = b"SETNX"  # Deprecated in redis: 2.6.12
    SUBSTR = b"SUBSTR"  # Deprecated in redis: 2.0.0
    SETEX = b"SETEX"  # Deprecated in redis: 2.6.12
    PSETEX = b"PSETEX"  # Deprecated in redis: 2.6.12

    #: Commands for cluster
    ASKING = b"ASKING"  # Since redis: 3.0.0
    CLUSTER = b"CLUSTER"  # Since redis: 3.0.0
    CLUSTER_ADDSLOTS = b"CLUSTER ADDSLOTS"  # Since redis: 3.0.0
    CLUSTER_BUMPEPOCH = b"CLUSTER BUMPEPOCH"  # Since redis: 3.0.0
    CLUSTER_COUNT_FAILURE_REPORTS = (
        b"CLUSTER COUNT-FAILURE-REPORTS"  # Since redis: 3.0.0
    )
    CLUSTER_COUNTKEYSINSLOT = b"CLUSTER COUNTKEYSINSLOT"  # Since redis: 3.0.0
    CLUSTER_DELSLOTS = b"CLUSTER DELSLOTS"  # Since redis: 3.0.0
    CLUSTER_FAILOVER = b"CLUSTER FAILOVER"  # Since redis: 3.0.0
    CLUSTER_FLUSHSLOTS = b"CLUSTER FLUSHSLOTS"  # Since redis: 3.0.0
    CLUSTER_FORGET = b"CLUSTER FORGET"  # Since redis: 3.0.0
    CLUSTER_GETKEYSINSLOT = b"CLUSTER GETKEYSINSLOT"  # Since redis: 3.0.0
    CLUSTER_INFO = b"CLUSTER INFO"  # Since redis: 3.0.0
    CLUSTER_KEYSLOT = b"CLUSTER KEYSLOT"  # Since redis: 3.0.0
    CLUSTER_MEET = b"CLUSTER MEET"  # Since redis: 3.0.0
    CLUSTER_MYID = b"CLUSTER MYID"  # Since redis: 3.0.0
    CLUSTER_NODES = b"CLUSTER NODES"  # Since redis: 3.0.0
    CLUSTER_REPLICATE = b"CLUSTER REPLICATE"  # Since redis: 3.0.0
    CLUSTER_RESET = b"CLUSTER RESET"  # Since redis: 3.0.0
    CLUSTER_SAVECONFIG = b"CLUSTER SAVECONFIG"  # Since redis: 3.0.0
    CLUSTER_SET_CONFIG_EPOCH = b"CLUSTER SET-CONFIG-EPOCH"  # Since redis: 3.0.0
    CLUSTER_SETSLOT = b"CLUSTER SETSLOT"  # Since redis: 3.0.0
    READONLY = b"READONLY"  # Since redis: 3.0.0
    READWRITE = b"READWRITE"  # Since redis: 3.0.0
    CLUSTER_HELP = b"CLUSTER HELP"  # Since redis: 5.0.0
    CLUSTER_REPLICAS = b"CLUSTER REPLICAS"  # Since redis: 5.0.0
    CLUSTER_ADDSLOTSRANGE = b"CLUSTER ADDSLOTSRANGE"  # Since redis: 7.0.0
    CLUSTER_DELSLOTSRANGE = b"CLUSTER DELSLOTSRANGE"  # Since redis: 7.0.0
    CLUSTER_LINKS = b"CLUSTER LINKS"  # Since redis: 7.0.0
    CLUSTER_SHARDS = b"CLUSTER SHARDS"  # Since redis: 7.0.0
    CLUSTER_MYSHARDID = b"CLUSTER MYSHARDID"  # Since redis: 7.2.0
    CLUSTER_SLAVES = b"CLUSTER SLAVES"  # Deprecated in redis: 5.0.0
    CLUSTER_SLOTS = b"CLUSTER SLOTS"  # Deprecated in redis: 7.0.0

    #: Commands for connection
    AUTH = b"AUTH"  # Since redis: 1.0.0
    ECHO = b"ECHO"  # Since redis: 1.0.0
    PING = b"PING"  # Since redis: 1.0.0
    SELECT = b"SELECT"  # Since redis: 1.0.0
    CLIENT = b"CLIENT"  # Since redis: 2.4.0
    CLIENT_KILL = b"CLIENT KILL"  # Since redis: 2.4.0
    CLIENT_LIST = b"CLIENT LIST"  # Since redis: 2.4.0
    CLIENT_GETNAME = b"CLIENT GETNAME"  # Since redis: 2.6.9
    CLIENT_SETNAME = b"CLIENT SETNAME"  # Since redis: 2.6.9
    CLIENT_PAUSE = b"CLIENT PAUSE"  # Since redis: 3.0.0
    CLIENT_REPLY = b"CLIENT REPLY"  # Since redis: 3.2.0
    CLIENT_HELP = b"CLIENT HELP"  # Since redis: 5.0.0
    CLIENT_ID = b"CLIENT ID"  # Since redis: 5.0.0
    CLIENT_UNBLOCK = b"CLIENT UNBLOCK"  # Since redis: 5.0.0
    CLIENT_CACHING = b"CLIENT CACHING"  # Since redis: 6.0.0
    CLIENT_GETREDIR = b"CLIENT GETREDIR"  # Since redis: 6.0.0
    CLIENT_TRACKING = b"CLIENT TRACKING"  # Since redis: 6.0.0
    HELLO = b"HELLO"  # Since redis: 6.0.0
    CLIENT_INFO = b"CLIENT INFO"  # Since redis: 6.2.0
    CLIENT_TRACKINGINFO = b"CLIENT TRACKINGINFO"  # Since redis: 6.2.0
    CLIENT_UNPAUSE = b"CLIENT UNPAUSE"  # Since redis: 6.2.0
    RESET = b"RESET"  # Since redis: 6.2.0
    CLIENT_NO_EVICT = b"CLIENT NO-EVICT"  # Since redis: 7.0.0
    CLIENT_NO_TOUCH = b"CLIENT NO-TOUCH"  # Since redis: 7.2.0
    CLIENT_SETINFO = b"CLIENT SETINFO"  # Since redis: 7.2.0
    QUIT = b"QUIT"  # Deprecated in redis: 7.2.0

    #: Commands for bitmap
    GETBIT = b"GETBIT"  # Since redis: 2.2.0
    SETBIT = b"SETBIT"  # Since redis: 2.2.0
    BITCOUNT = b"BITCOUNT"  # Since redis: 2.6.0
    BITOP = b"BITOP"  # Since redis: 2.6.0
    BITPOS = b"BITPOS"  # Since redis: 2.8.7
    BITFIELD = b"BITFIELD"  # Since redis: 3.2.0
    BITFIELD_RO = b"BITFIELD_RO"  # Since redis: 6.0.0

    #: Commands for list
    LINDEX = b"LINDEX"  # Since redis: 1.0.0
    LLEN = b"LLEN"  # Since redis: 1.0.0
    LPOP = b"LPOP"  # Since redis: 1.0.0
    LPUSH = b"LPUSH"  # Since redis: 1.0.0
    LRANGE = b"LRANGE"  # Since redis: 1.0.0
    LREM = b"LREM"  # Since redis: 1.0.0
    LSET = b"LSET"  # Since redis: 1.0.0
    LTRIM = b"LTRIM"  # Since redis: 1.0.0
    RPOP = b"RPOP"  # Since redis: 1.0.0
    RPUSH = b"RPUSH"  # Since redis: 1.0.0
    BLPOP = b"BLPOP"  # Since redis: 2.0.0
    BRPOP = b"BRPOP"  # Since redis: 2.0.0
    LINSERT = b"LINSERT"  # Since redis: 2.2.0
    LPUSHX = b"LPUSHX"  # Since redis: 2.2.0
    RPUSHX = b"RPUSHX"  # Since redis: 2.2.0
    LPOS = b"LPOS"  # Since redis: 6.0.6
    BLMOVE = b"BLMOVE"  # Since redis: 6.2.0
    LMOVE = b"LMOVE"  # Since redis: 6.2.0
    BLMPOP = b"BLMPOP"  # Since redis: 7.0.0
    LMPOP = b"LMPOP"  # Since redis: 7.0.0
    RPOPLPUSH = b"RPOPLPUSH"  # Deprecated in redis: 6.2.0
    BRPOPLPUSH = b"BRPOPLPUSH"  # Deprecated in redis: 6.2.0

    #: Commands for sorted-set
    ZADD = b"ZADD"  # Since redis: 1.2.0
    ZCARD = b"ZCARD"  # Since redis: 1.2.0
    ZINCRBY = b"ZINCRBY"  # Since redis: 1.2.0
    ZRANGE = b"ZRANGE"  # Since redis: 1.2.0
    ZREM = b"ZREM"  # Since redis: 1.2.0
    ZREMRANGEBYSCORE = b"ZREMRANGEBYSCORE"  # Since redis: 1.2.0
    ZSCORE = b"ZSCORE"  # Since redis: 1.2.0
    ZCOUNT = b"ZCOUNT"  # Since redis: 2.0.0
    ZINTERSTORE = b"ZINTERSTORE"  # Since redis: 2.0.0
    ZRANK = b"ZRANK"  # Since redis: 2.0.0
    ZREMRANGEBYRANK = b"ZREMRANGEBYRANK"  # Since redis: 2.0.0
    ZREVRANK = b"ZREVRANK"  # Since redis: 2.0.0
    ZUNIONSTORE = b"ZUNIONSTORE"  # Since redis: 2.0.0
    ZSCAN = b"ZSCAN"  # Since redis: 2.8.0
    ZLEXCOUNT = b"ZLEXCOUNT"  # Since redis: 2.8.9
    ZREMRANGEBYLEX = b"ZREMRANGEBYLEX"  # Since redis: 2.8.9
    BZPOPMAX = b"BZPOPMAX"  # Since redis: 5.0.0
    BZPOPMIN = b"BZPOPMIN"  # Since redis: 5.0.0
    ZPOPMAX = b"ZPOPMAX"  # Since redis: 5.0.0
    ZPOPMIN = b"ZPOPMIN"  # Since redis: 5.0.0
    ZDIFF = b"ZDIFF"  # Since redis: 6.2.0
    ZDIFFSTORE = b"ZDIFFSTORE"  # Since redis: 6.2.0
    ZINTER = b"ZINTER"  # Since redis: 6.2.0
    ZMSCORE = b"ZMSCORE"  # Since redis: 6.2.0
    ZRANDMEMBER = b"ZRANDMEMBER"  # Since redis: 6.2.0
    ZRANGESTORE = b"ZRANGESTORE"  # Since redis: 6.2.0
    ZUNION = b"ZUNION"  # Since redis: 6.2.0
    BZMPOP = b"BZMPOP"  # Since redis: 7.0.0
    ZINTERCARD = b"ZINTERCARD"  # Since redis: 7.0.0
    ZMPOP = b"ZMPOP"  # Since redis: 7.0.0
    ZRANGEBYSCORE = b"ZRANGEBYSCORE"  # Deprecated in redis: 6.2.0
    ZREVRANGE = b"ZREVRANGE"  # Deprecated in redis: 6.2.0
    ZREVRANGEBYSCORE = b"ZREVRANGEBYSCORE"  # Deprecated in redis: 6.2.0
    ZRANGEBYLEX = b"ZRANGEBYLEX"  # Deprecated in redis: 6.2.0
    ZREVRANGEBYLEX = b"ZREVRANGEBYLEX"  # Deprecated in redis: 6.2.0

    #: Commands for generic
    DEL = b"DEL"  # Since redis: 1.0.0
    EXISTS = b"EXISTS"  # Since redis: 1.0.0
    EXPIRE = b"EXPIRE"  # Since redis: 1.0.0
    KEYS = b"KEYS"  # Since redis: 1.0.0
    MOVE = b"MOVE"  # Since redis: 1.0.0
    RANDOMKEY = b"RANDOMKEY"  # Since redis: 1.0.0
    RENAME = b"RENAME"  # Since redis: 1.0.0
    RENAMENX = b"RENAMENX"  # Since redis: 1.0.0
    SORT = b"SORT"  # Since redis: 1.0.0
    TTL = b"TTL"  # Since redis: 1.0.0
    TYPE = b"TYPE"  # Since redis: 1.0.0
    EXPIREAT = b"EXPIREAT"  # Since redis: 1.2.0
    PERSIST = b"PERSIST"  # Since redis: 2.2.0
    OBJECT = b"OBJECT"  # Since redis: 2.2.3
    OBJECT_ENCODING = b"OBJECT ENCODING"  # Since redis: 2.2.3
    OBJECT_IDLETIME = b"OBJECT IDLETIME"  # Since redis: 2.2.3
    OBJECT_REFCOUNT = b"OBJECT REFCOUNT"  # Since redis: 2.2.3
    DUMP = b"DUMP"  # Since redis: 2.6.0
    MIGRATE = b"MIGRATE"  # Since redis: 2.6.0
    PEXPIRE = b"PEXPIRE"  # Since redis: 2.6.0
    PEXPIREAT = b"PEXPIREAT"  # Since redis: 2.6.0
    PTTL = b"PTTL"  # Since redis: 2.6.0
    RESTORE = b"RESTORE"  # Since redis: 2.6.0
    SCAN = b"SCAN"  # Since redis: 2.8.0
    WAIT = b"WAIT"  # Since redis: 3.0.0
    TOUCH = b"TOUCH"  # Since redis: 3.2.1
    OBJECT_FREQ = b"OBJECT FREQ"  # Since redis: 4.0.0
    UNLINK = b"UNLINK"  # Since redis: 4.0.0
    COPY = b"COPY"  # Since redis: 6.2.0
    OBJECT_HELP = b"OBJECT HELP"  # Since redis: 6.2.0
    EXPIRETIME = b"EXPIRETIME"  # Since redis: 7.0.0
    PEXPIRETIME = b"PEXPIRETIME"  # Since redis: 7.0.0
    SORT_RO = b"SORT_RO"  # Since redis: 7.0.0
    WAITAOF = b"WAITAOF"  # Since redis: 7.2.0

    #: Commands for transactions
    EXEC = b"EXEC"  # Since redis: 1.2.0
    MULTI = b"MULTI"  # Since redis: 1.2.0
    DISCARD = b"DISCARD"  # Since redis: 2.0.0
    UNWATCH = b"UNWATCH"  # Since redis: 2.2.0
    WATCH = b"WATCH"  # Since redis: 2.2.0

    #: Commands for scripting
    EVAL = b"EVAL"  # Since redis: 2.6.0
    EVALSHA = b"EVALSHA"  # Since redis: 2.6.0
    SCRIPT = b"SCRIPT"  # Since redis: 2.6.0
    SCRIPT_EXISTS = b"SCRIPT EXISTS"  # Since redis: 2.6.0
    SCRIPT_FLUSH = b"SCRIPT FLUSH"  # Since redis: 2.6.0
    SCRIPT_KILL = b"SCRIPT KILL"  # Since redis: 2.6.0
    SCRIPT_LOAD = b"SCRIPT LOAD"  # Since redis: 2.6.0
    SCRIPT_DEBUG = b"SCRIPT DEBUG"  # Since redis: 3.2.0
    SCRIPT_HELP = b"SCRIPT HELP"  # Since redis: 5.0.0
    EVALSHA_RO = b"EVALSHA_RO"  # Since redis: 7.0.0
    EVAL_RO = b"EVAL_RO"  # Since redis: 7.0.0
    FCALL = b"FCALL"  # Since redis: 7.0.0
    FCALL_RO = b"FCALL_RO"  # Since redis: 7.0.0
    FUNCTION = b"FUNCTION"  # Since redis: 7.0.0
    FUNCTION_DELETE = b"FUNCTION DELETE"  # Since redis: 7.0.0
    FUNCTION_DUMP = b"FUNCTION DUMP"  # Since redis: 7.0.0
    FUNCTION_FLUSH = b"FUNCTION FLUSH"  # Since redis: 7.0.0
    FUNCTION_HELP = b"FUNCTION HELP"  # Since redis: 7.0.0
    FUNCTION_KILL = b"FUNCTION KILL"  # Since redis: 7.0.0
    FUNCTION_LIST = b"FUNCTION LIST"  # Since redis: 7.0.0
    FUNCTION_LOAD = b"FUNCTION LOAD"  # Since redis: 7.0.0
    FUNCTION_RESTORE = b"FUNCTION RESTORE"  # Since redis: 7.0.0
    FUNCTION_STATS = b"FUNCTION STATS"  # Since redis: 7.0.0

    #: Commands for geo
    GEOADD = b"GEOADD"  # Since redis: 3.2.0
    GEODIST = b"GEODIST"  # Since redis: 3.2.0
    GEOHASH = b"GEOHASH"  # Since redis: 3.2.0
    GEOPOS = b"GEOPOS"  # Since redis: 3.2.0
    GEOSEARCH = b"GEOSEARCH"  # Since redis: 6.2.0
    GEOSEARCHSTORE = b"GEOSEARCHSTORE"  # Since redis: 6.2.0
    GEORADIUS = b"GEORADIUS"  # Deprecated in redis: 6.2.0
    GEORADIUSBYMEMBER = b"GEORADIUSBYMEMBER"  # Deprecated in redis: 6.2.0
    GEORADIUSBYMEMBER_RO = b"GEORADIUSBYMEMBER_RO"  # Deprecated in redis: 6.2.0
    GEORADIUS_RO = b"GEORADIUS_RO"  # Deprecated in redis: 6.2.0

    #: Commands for hash
    HDEL = b"HDEL"  # Since redis: 2.0.0
    HEXISTS = b"HEXISTS"  # Since redis: 2.0.0
    HGET = b"HGET"  # Since redis: 2.0.0
    HGETALL = b"HGETALL"  # Since redis: 2.0.0
    HINCRBY = b"HINCRBY"  # Since redis: 2.0.0
    HKEYS = b"HKEYS"  # Since redis: 2.0.0
    HLEN = b"HLEN"  # Since redis: 2.0.0
    HMGET = b"HMGET"  # Since redis: 2.0.0
    HSET = b"HSET"  # Since redis: 2.0.0
    HSETNX = b"HSETNX"  # Since redis: 2.0.0
    HVALS = b"HVALS"  # Since redis: 2.0.0
    HINCRBYFLOAT = b"HINCRBYFLOAT"  # Since redis: 2.6.0
    HSCAN = b"HSCAN"  # Since redis: 2.8.0
    HSTRLEN = b"HSTRLEN"  # Since redis: 3.2.0
    HRANDFIELD = b"HRANDFIELD"  # Since redis: 6.2.0
    HMSET = b"HMSET"  # Deprecated in redis: 4.0.0

    #: Commands for hyperloglog
    PFADD = b"PFADD"  # Since redis: 2.8.9
    PFCOUNT = b"PFCOUNT"  # Since redis: 2.8.9
    PFDEBUG = b"PFDEBUG"  # Since redis: 2.8.9
    PFMERGE = b"PFMERGE"  # Since redis: 2.8.9
    PFSELFTEST = b"PFSELFTEST"  # Since redis: 2.8.9

    #: Commands for pubsub
    PSUBSCRIBE = b"PSUBSCRIBE"  # Since redis: 2.0.0
    PUBLISH = b"PUBLISH"  # Since redis: 2.0.0
    PUNSUBSCRIBE = b"PUNSUBSCRIBE"  # Since redis: 2.0.0
    SUBSCRIBE = b"SUBSCRIBE"  # Since redis: 2.0.0
    UNSUBSCRIBE = b"UNSUBSCRIBE"  # Since redis: 2.0.0
    PUBSUB = b"PUBSUB"  # Since redis: 2.8.0
    PUBSUB_CHANNELS = b"PUBSUB CHANNELS"  # Since redis: 2.8.0
    PUBSUB_NUMPAT = b"PUBSUB NUMPAT"  # Since redis: 2.8.0
    PUBSUB_NUMSUB = b"PUBSUB NUMSUB"  # Since redis: 2.8.0
    PUBSUB_HELP = b"PUBSUB HELP"  # Since redis: 6.2.0
    PUBSUB_SHARDCHANNELS = b"PUBSUB SHARDCHANNELS"  # Since redis: 7.0.0
    PUBSUB_SHARDNUMSUB = b"PUBSUB SHARDNUMSUB"  # Since redis: 7.0.0
    SPUBLISH = b"SPUBLISH"  # Since redis: 7.0.0
    SSUBSCRIBE = b"SSUBSCRIBE"  # Since redis: 7.0.0
    SUNSUBSCRIBE = b"SUNSUBSCRIBE"  # Since redis: 7.0.0

    #: Commands for set
    SADD = b"SADD"  # Since redis: 1.0.0
    SCARD = b"SCARD"  # Since redis: 1.0.0
    SDIFF = b"SDIFF"  # Since redis: 1.0.0
    SDIFFSTORE = b"SDIFFSTORE"  # Since redis: 1.0.0
    SINTER = b"SINTER"  # Since redis: 1.0.0
    SINTERSTORE = b"SINTERSTORE"  # Since redis: 1.0.0
    SISMEMBER = b"SISMEMBER"  # Since redis: 1.0.0
    SMEMBERS = b"SMEMBERS"  # Since redis: 1.0.0
    SMOVE = b"SMOVE"  # Since redis: 1.0.0
    SPOP = b"SPOP"  # Since redis: 1.0.0
    SRANDMEMBER = b"SRANDMEMBER"  # Since redis: 1.0.0
    SREM = b"SREM"  # Since redis: 1.0.0
    SUNION = b"SUNION"  # Since redis: 1.0.0
    SUNIONSTORE = b"SUNIONSTORE"  # Since redis: 1.0.0
    SSCAN = b"SSCAN"  # Since redis: 2.8.0
    SMISMEMBER = b"SMISMEMBER"  # Since redis: 6.2.0
    SINTERCARD = b"SINTERCARD"  # Since redis: 7.0.0

    #: Commands for stream
    XACK = b"XACK"  # Since redis: 5.0.0
    XADD = b"XADD"  # Since redis: 5.0.0
    XCLAIM = b"XCLAIM"  # Since redis: 5.0.0
    XDEL = b"XDEL"  # Since redis: 5.0.0
    XGROUP = b"XGROUP"  # Since redis: 5.0.0
    XGROUP_CREATE = b"XGROUP CREATE"  # Since redis: 5.0.0
    XGROUP_DELCONSUMER = b"XGROUP DELCONSUMER"  # Since redis: 5.0.0
    XGROUP_DESTROY = b"XGROUP DESTROY"  # Since redis: 5.0.0
    XGROUP_HELP = b"XGROUP HELP"  # Since redis: 5.0.0
    XGROUP_SETID = b"XGROUP SETID"  # Since redis: 5.0.0
    XINFO = b"XINFO"  # Since redis: 5.0.0
    XINFO_CONSUMERS = b"XINFO CONSUMERS"  # Since redis: 5.0.0
    XINFO_GROUPS = b"XINFO GROUPS"  # Since redis: 5.0.0
    XINFO_HELP = b"XINFO HELP"  # Since redis: 5.0.0
    XINFO_STREAM = b"XINFO STREAM"  # Since redis: 5.0.0
    XLEN = b"XLEN"  # Since redis: 5.0.0
    XPENDING = b"XPENDING"  # Since redis: 5.0.0
    XRANGE = b"XRANGE"  # Since redis: 5.0.0
    XREAD = b"XREAD"  # Since redis: 5.0.0
    XREADGROUP = b"XREADGROUP"  # Since redis: 5.0.0
    XREVRANGE = b"XREVRANGE"  # Since redis: 5.0.0
    XSETID = b"XSETID"  # Since redis: 5.0.0
    XTRIM = b"XTRIM"  # Since redis: 5.0.0
    XAUTOCLAIM = b"XAUTOCLAIM"  # Since redis: 6.2.0
    XGROUP_CREATECONSUMER = b"XGROUP CREATECONSUMER"  # Since redis: 6.2.0

    #: Commands for json
    JSON_DEL = b"JSON.DEL"  # Since RedisJSON: 1.0.0
    JSON_FORGET = b"JSON.FORGET"  # Since RedisJSON: 1.0.0
    JSON_GET = b"JSON.GET"  # Since RedisJSON: 1.0.0
    JSON_SET = b"JSON.SET"  # Since RedisJSON: 1.0.0
    JSON_MGET = b"JSON.MGET"  # Since RedisJSON: 1.0.0
    JSON_NUMINCRBY = b"JSON.NUMINCRBY"  # Since RedisJSON: 1.0.0
    JSON_STRAPPEND = b"JSON.STRAPPEND"  # Since RedisJSON: 1.0.0
    JSON_STRLEN = b"JSON.STRLEN"  # Since RedisJSON: 1.0.0
    JSON_ARRAPPEND = b"JSON.ARRAPPEND"  # Since RedisJSON: 1.0.0
    JSON_ARRINDEX = b"JSON.ARRINDEX"  # Since RedisJSON: 1.0.0
    JSON_ARRINSERT = b"JSON.ARRINSERT"  # Since RedisJSON: 1.0.0
    JSON_ARRLEN = b"JSON.ARRLEN"  # Since RedisJSON: 1.0.0
    JSON_ARRPOP = b"JSON.ARRPOP"  # Since RedisJSON: 1.0.0
    JSON_ARRTRIM = b"JSON.ARRTRIM"  # Since RedisJSON: 1.0.0
    JSON_OBJKEYS = b"JSON.OBJKEYS"  # Since RedisJSON: 1.0.0
    JSON_OBJLEN = b"JSON.OBJLEN"  # Since RedisJSON: 1.0.0
    JSON_TYPE = b"JSON.TYPE"  # Since RedisJSON: 1.0.0
    JSON_RESP = b"JSON.RESP"  # Since RedisJSON: 1.0.0
    JSON_DEBUG = b"JSON.DEBUG"  # Since RedisJSON: 1.0.0
    JSON_DEBUG_HELP = b"JSON.DEBUG HELP"  # Since RedisJSON: 1.0.0
    JSON_DEBUG_MEMORY = b"JSON.DEBUG MEMORY"  # Since RedisJSON: 1.0.0
    JSON_TOGGLE = b"JSON.TOGGLE"  # Since RedisJSON: 2.0.0
    JSON_CLEAR = b"JSON.CLEAR"  # Since RedisJSON: 2.0.0
    JSON_MSET = b"JSON.MSET"  # Since RedisJSON: 2.6.0
    JSON_MERGE = b"JSON.MERGE"  # Since RedisJSON: 2.6.0
    JSON_NUMMULTBY = b"JSON.NUMMULTBY"  # Deprecated in RedisJSON: 2.0

    #: Commands for bf
    BF_RESERVE = b"BF.RESERVE"  # Since bf: 1.0.0
    BF_ADD = b"BF.ADD"  # Since bf: 1.0.0
    BF_MADD = b"BF.MADD"  # Since bf: 1.0.0
    BF_INSERT = b"BF.INSERT"  # Since bf: 1.0.0
    BF_EXISTS = b"BF.EXISTS"  # Since bf: 1.0.0
    BF_MEXISTS = b"BF.MEXISTS"  # Since bf: 1.0.0
    BF_SCANDUMP = b"BF.SCANDUMP"  # Since bf: 1.0.0
    BF_LOADCHUNK = b"BF.LOADCHUNK"  # Since bf: 1.0.0
    BF_INFO = b"BF.INFO"  # Since bf: 1.0.0
    BF_CARD = b"BF.CARD"  # Since bf: 2.4.4

    #: Commands for cf
    CF_RESERVE = b"CF.RESERVE"  # Since bf: 1.0.0
    CF_ADD = b"CF.ADD"  # Since bf: 1.0.0
    CF_ADDNX = b"CF.ADDNX"  # Since bf: 1.0.0
    CF_INSERT = b"CF.INSERT"  # Since bf: 1.0.0
    CF_INSERTNX = b"CF.INSERTNX"  # Since bf: 1.0.0
    CF_EXISTS = b"CF.EXISTS"  # Since bf: 1.0.0
    CF_MEXISTS = b"CF.MEXISTS"  # Since bf: 1.0.0
    CF_DEL = b"CF.DEL"  # Since bf: 1.0.0
    CF_COUNT = b"CF.COUNT"  # Since bf: 1.0.0
    CF_SCANDUMP = b"CF.SCANDUMP"  # Since bf: 1.0.0
    CF_LOADCHUNK = b"CF.LOADCHUNK"  # Since bf: 1.0.0
    CF_INFO = b"CF.INFO"  # Since bf: 1.0.0

    #: Commands for cms
    CMS_INITBYDIM = b"CMS.INITBYDIM"  # Since bf: 2.0.0
    CMS_INITBYPROB = b"CMS.INITBYPROB"  # Since bf: 2.0.0
    CMS_INCRBY = b"CMS.INCRBY"  # Since bf: 2.0.0
    CMS_QUERY = b"CMS.QUERY"  # Since bf: 2.0.0
    CMS_MERGE = b"CMS.MERGE"  # Since bf: 2.0.0
    CMS_INFO = b"CMS.INFO"  # Since bf: 2.0.0

    #: Commands for topk
    TOPK_RESERVE = b"TOPK.RESERVE"  # Since bf: 2.0.0
    TOPK_ADD = b"TOPK.ADD"  # Since bf: 2.0.0
    TOPK_INCRBY = b"TOPK.INCRBY"  # Since bf: 2.0.0
    TOPK_QUERY = b"TOPK.QUERY"  # Since bf: 2.0.0
    TOPK_LIST = b"TOPK.LIST"  # Since bf: 2.0.0
    TOPK_INFO = b"TOPK.INFO"  # Since bf: 2.0.0
    TOPK_COUNT = b"TOPK.COUNT"  # Deprecated in bf: 2.4

    #: Commands for tdigest
    TDIGEST_CREATE = b"TDIGEST.CREATE"  # Since bf: 2.4.0
    TDIGEST_RESET = b"TDIGEST.RESET"  # Since bf: 2.4.0
    TDIGEST_ADD = b"TDIGEST.ADD"  # Since bf: 2.4.0
    TDIGEST_MERGE = b"TDIGEST.MERGE"  # Since bf: 2.4.0
    TDIGEST_MIN = b"TDIGEST.MIN"  # Since bf: 2.4.0
    TDIGEST_MAX = b"TDIGEST.MAX"  # Since bf: 2.4.0
    TDIGEST_QUANTILE = b"TDIGEST.QUANTILE"  # Since bf: 2.4.0
    TDIGEST_CDF = b"TDIGEST.CDF"  # Since bf: 2.4.0
    TDIGEST_TRIMMED_MEAN = b"TDIGEST.TRIMMED_MEAN"  # Since bf: 2.4.0
    TDIGEST_RANK = b"TDIGEST.RANK"  # Since bf: 2.4.0
    TDIGEST_REVRANK = b"TDIGEST.REVRANK"  # Since bf: 2.4.0
    TDIGEST_BYRANK = b"TDIGEST.BYRANK"  # Since bf: 2.4.0
    TDIGEST_BYREVRANK = b"TDIGEST.BYREVRANK"  # Since bf: 2.4.0
    TDIGEST_INFO = b"TDIGEST.INFO"  # Since bf: 2.4.0

    #: Commands for timeseries
    TS_CREATE = b"TS.CREATE"  # Since timeseries: 1.0.0
    TS_ALTER = b"TS.ALTER"  # Since timeseries: 1.0.0
    TS_ADD = b"TS.ADD"  # Since timeseries: 1.0.0
    TS_MADD = b"TS.MADD"  # Since timeseries: 1.0.0
    TS_INCRBY = b"TS.INCRBY"  # Since timeseries: 1.0.0
    TS_DECRBY = b"TS.DECRBY"  # Since timeseries: 1.0.0
    TS_CREATERULE = b"TS.CREATERULE"  # Since timeseries: 1.0.0
    TS_DELETERULE = b"TS.DELETERULE"  # Since timeseries: 1.0.0
    TS_RANGE = b"TS.RANGE"  # Since timeseries: 1.0.0
    TS_MRANGE = b"TS.MRANGE"  # Since timeseries: 1.0.0
    TS_GET = b"TS.GET"  # Since timeseries: 1.0.0
    TS_MGET = b"TS.MGET"  # Since timeseries: 1.0.0
    TS_INFO = b"TS.INFO"  # Since timeseries: 1.0.0
    TS_QUERYINDEX = b"TS.QUERYINDEX"  # Since timeseries: 1.0.0
    TS_REVRANGE = b"TS.REVRANGE"  # Since timeseries: 1.4.0
    TS_MREVRANGE = b"TS.MREVRANGE"  # Since timeseries: 1.4.0
    TS_DEL = b"TS.DEL"  # Since timeseries: 1.6.0

    #: Commands for graph
    GRAPH_QUERY = b"GRAPH.QUERY"  # Since graph: 1.0.0
    GRAPH_DELETE = b"GRAPH.DELETE"  # Since graph: 1.0.0
    GRAPH_EXPLAIN = b"GRAPH.EXPLAIN"  # Since graph: 2.0.0
    GRAPH_PROFILE = b"GRAPH.PROFILE"  # Since graph: 2.0.0
    GRAPH_SLOWLOG = b"GRAPH.SLOWLOG"  # Since graph: 2.0.12
    GRAPH_CONSTRAINT_DROP = b"GRAPH.CONSTRAINT DROP"  # Since graph: 2.12.0
    GRAPH_CONSTRAINT_CREATE = b"GRAPH.CONSTRAINT CREATE"  # Since graph: 2.12.0
    GRAPH_CONFIG_GET = b"GRAPH.CONFIG GET"  # Since graph: 2.2.11
    GRAPH_CONFIG_SET = b"GRAPH.CONFIG SET"  # Since graph: 2.2.11
    GRAPH_RO_QUERY = b"GRAPH.RO_QUERY"  # Since graph: 2.2.8
    GRAPH_LIST = b"GRAPH.LIST"  # Since graph: 2.4.3

    #: Commands for search
    FT_CREATE = b"FT.CREATE"  # Since search: 1.0.0
    FT_INFO = b"FT.INFO"  # Since search: 1.0.0
    FT_EXPLAIN = b"FT.EXPLAIN"  # Since search: 1.0.0
    FT_EXPLAINCLI = b"FT.EXPLAINCLI"  # Since search: 1.0.0
    FT_ALTER = b"FT.ALTER"  # Since search: 1.0.0
    FT_ALIASADD = b"FT.ALIASADD"  # Since search: 1.0.0
    FT_ALIASUPDATE = b"FT.ALIASUPDATE"  # Since search: 1.0.0
    FT_ALIASDEL = b"FT.ALIASDEL"  # Since search: 1.0.0
    FT_TAGVALS = b"FT.TAGVALS"  # Since search: 1.0.0
    FT_CONFIG_SET = b"FT.CONFIG SET"  # Since search: 1.0.0
    FT_CONFIG_GET = b"FT.CONFIG GET"  # Since search: 1.0.0
    FT_CONFIG_HELP = b"FT.CONFIG HELP"  # Since search: 1.0.0
    FT_SEARCH = b"FT.SEARCH"  # Since search: 1.0.0
    FT_AGGREGATE = b"FT.AGGREGATE"  # Since search: 1.1.0
    FT_CURSOR_READ = b"FT.CURSOR READ"  # Since search: 1.1.0
    FT_CURSOR_DEL = b"FT.CURSOR DEL"  # Since search: 1.1.0
    FT_SYNUPDATE = b"FT.SYNUPDATE"  # Since search: 1.2.0
    FT_SYNDUMP = b"FT.SYNDUMP"  # Since search: 1.2.0
    FT_SPELLCHECK = b"FT.SPELLCHECK"  # Since search: 1.4.0
    FT_DICTADD = b"FT.DICTADD"  # Since search: 1.4.0
    FT_DICTDEL = b"FT.DICTDEL"  # Since search: 1.4.0
    FT_DICTDUMP = b"FT.DICTDUMP"  # Since search: 1.4.0
    FT_DROPINDEX = b"FT.DROPINDEX"  # Since search: 2.0.0
    FT__LIST = b"FT._LIST"  # Since search: 2.0.0
    FT_PROFILE = b"FT.PROFILE"  # Since search: 2.2.0

    #: Commands for suggestion
    FT_SUGADD = b"FT.SUGADD"  # Since search: 1.0.0
    FT_SUGGET = b"FT.SUGGET"  # Since search: 1.0.0
    FT_SUGDEL = b"FT.SUGDEL"  # Since search: 1.0.0
    FT_SUGLEN = b"FT.SUGLEN"  # Since search: 1.0.0

    #: Oddball command
    DEBUG_OBJECT = b"DEBUG OBJECT"

    #: Sentinel commands
    SENTINEL_CKQUORUM = b"SENTINEL CKQUORUM"
    SENTINEL_CONFIG_GET = b"SENTINEL CONFIG GET"
    SENTINEL_CONFIG_SET = b"SENTINEL CONFIG SET"
    SENTINEL_GET_MASTER_ADDR_BY_NAME = b"SENTINEL GET-MASTER-ADDR-BY-NAME"
    SENTINEL_FAILOVER = b"SENTINEL FAILOVER"
    SENTINEL_FLUSHCONFIG = b"SENTINEL FLUSHCONFIG"
    SENTINEL_INFO_CACHE = b"SENTINEL INFO-CACHE"
    SENTINEL_IS_MASTER_DOWN_BY_ADDR = b"SENTINEL IS-MASTER-DOWN-BY-ADDR"
    SENTINEL_MASTER = b"SENTINEL MASTER"
    SENTINEL_MASTERS = b"SENTINEL MASTERS"
    SENTINEL_MONITOR = b"SENTINEL MONITOR"
    SENTINEL_MYID = b"SENTINEL MYID"
    SENTINEL_PENDING_SCRIPTS = b"SENTINEL PENDING-SCRIPTS"
    SENTINEL_REMOVE = b"SENTINEL REMOVE"
    SENTINEL_SLAVES = b"SENTINEL SLAVES"  # Deprecated
    SENTINEL_REPLICAS = b"SENTINEL REPLICAS"
    SENTINEL_RESET = b"SENTINEL RESET"
    SENTINEL_SENTINELS = b"SENTINEL SENTINELS"
    SENTINEL_SET = b"SENTINEL SET"


class CommandGroup(enum.Enum):
    BF = "bf"
    BITMAP = "bitmap"
    CF = "cf"
    CLUSTER = "cluster"
    CMS = "cms"
    CONNECTION = "connection"
    GENERIC = "generic"
    GEO = "geo"
    GRAPH = "graph"
    HASH = "hash"
    HYPERLOGLOG = "hyperloglog"
    JSON = "json"
    LIST = "list"
    PUBSUB = "pubsub"
    SCRIPTING = "scripting"
    SEARCH = "search"
    SERVER = "server"
    SET = "set"
    SORTED_SET = "sorted-set"
    STREAM = "stream"
    STRING = "string"
    SUGGESTION = "suggestion"
    TDIGEST = "tdigest"
    TIMESERIES = "timeseries"
    TOPK = "topk"
    TRANSACTIONS = "transactions"


class NodeFlag(enum.Enum):
    ALL = "all"
    PRIMARIES = "primaries"
    REPLICAS = "replicas"
    RANDOM = "random"
    SLOT_ID = "slot-id"


class CommandFlag(enum.Enum):
    BLOCKING = "blocking"
    SLOW = "slow"
    FAST = "fast"
    READONLY = "readonly"



from __future__ import annotations

import functools
import hashlib
import inspect
import itertools
from typing import TYPE_CHECKING, Any, cast

from deprecated.sphinx import versionadded

from coredis._protocols import SupportsScript
from coredis._utils import b
from coredis.exceptions import NoScriptError
from coredis.typing import (
    AnyStr,
    Awaitable,
    Callable,
    Dict,
    Generic,
    KeyT,
    List,
    Optional,
    P,
    Parameters,
    R,
    ResponseType,
    StringT,
    Tuple,
    ValueT,
    add_runtime_checks,
    safe_beartype,
)

if TYPE_CHECKING:
    import coredis.client


class Script(Generic[AnyStr]):
    """
    An executable Lua script object returned by :meth:`coredis.Redis.register_script`.
    Instances of the class are callable and take arguments in the same shape
    as :meth:`coredis.Redis.evalsha` or :meth:`coredis.Redis.eval`
    (i.e a list of :paramref:`~__call__.keys` and :paramref:`~__call__.args`).

    Example::

        client = coredis.Redis()
        await client.set("test", "co")
        concat = client.register_script("return redis.call('GET', KEYS[1]) + ARGV[1]")
        assert await concat(['test'], ['redis']) == "coredis"
    """

    #: SHA of this script once it's registered with the redis server
    sha: AnyStr

    def __init__(
        self,
        registered_client: Optional[SupportsScript[AnyStr]] = None,
        script: Optional[StringT] = None,
        readonly: bool = False,
    ):
        """
        :param client: The client to use for executing the lua script. If
         this is ``None`` the client will have to be provided when invoking
         the script using :meth:`__call__` with the :paramref:`__call__.client`
         parameter.
        :param script: The lua script that will be used by :meth:`__call__`
        :param readonly: If ``True`` the script will be called with
         :meth:`coredis.Redis.evalsha_ro` instead of :meth:`coredis.Redis.evalsha`
        """
        self.registered_client: Optional[SupportsScript[AnyStr]] = registered_client
        self.script: StringT
        if not script:
            raise RuntimeError("No script provided")
        self.script = script
        self.sha = hashlib.sha1(b(script)).hexdigest()  # type: ignore
        self.readonly = readonly

    async def __call__(
        self,
        keys: Optional[Parameters[KeyT]] = None,
        args: Optional[Parameters[ValueT]] = None,
        client: Optional[SupportsScript[AnyStr]] = None,
        readonly: Optional[bool] = None,
    ) -> ResponseType:
        """
        Executes the script registered in :paramref:`Script.script` using
        :meth:`coredis.Redis.evalsha`. Additionally if the script was not yet
        registered on the instance, it will automatically do that as well
        and cache the sha at :data:`Script.sha`

        :param keys: The keys this script will reference
        :param args: The arguments expected by the script
        :param client: The redis client to use instead of :paramref:`Script.client`
        :param readonly: If ``True`` forces the script to be called with
         :meth:`coredis.Redis.evalsha_ro`
        """
        from coredis.pipeline import Pipeline

        if client is None:
            client = self.registered_client
        if not client:
            raise RuntimeError(
                "This instance is not bound to a redis client."
                "Please provide a valid instance to execute the script with"
            )
        if readonly is None:
            readonly = self.readonly

        # make sure the Redis server knows about the script
        if isinstance(client, Pipeline):
            # make sure this script is good to go on pipeline
            cast(Pipeline[AnyStr], client).scripts.add(self)

        method = client.evalsha_ro if readonly else client.evalsha
        try:
            return cast(ResponseType, await method(self.sha, keys=keys, args=args))
        except NoScriptError:
            # Maybe the client is pointed to a different server than the client
            # that created this instance?
            # Overwrite the sha just in case there was a discrepancy.
            self.sha = cast(AnyStr, await client.script_load(self.script))
            return cast(ResponseType, await method(self.sha, keys=keys, args=args))

    async def execute(
        self,
        keys: Optional[Parameters[KeyT]] = None,
        args: Optional[Parameters[ValueT]] = None,
        client: Optional[SupportsScript[AnyStr]] = None,
        readonly: Optional[bool] = None,
    ) -> ResponseType:
        """
        Executes the script registered in :paramref:`Script.script`

        :meta private:
        """
        return await self(keys, args, client, readonly)

    @versionadded(version="3.5.0")
    def wraps(
        self,
        key_spec: Optional[List[str]] = None,
        param_is_key: Callable[[inspect.Parameter], bool] = lambda p: (
            p.annotation in {"KeyT", KeyT}
        ),
        client_arg: Optional[str] = None,
        runtime_checks: bool = False,
        readonly: Optional[bool] = None,
    ) -> Callable[[Callable[P, Awaitable[R]]], Callable[P, Awaitable[R]]]:
        """
        Decorator for wrapping a regular python function, method or classmethod
        signature with a :class:`~coredis.commands.script.Script`. This allows
        exposing a strict signature instead of that which :meth:`__call__` provides.
        The callable being decorated should **not** have an implementation as
        it will never be called.

        The main objective of the decorator is to allow you to have strict (and type safe)
        signatures for wrappers for lua scripts. Internally the decorator separates
        ``keys`` from ``args`` before calling :meth:`coredis.Redis.evalsha`. Mapping the
        decorated methods arguments to key providers is done either by using :paramref:`key_spec`
        or :paramref:`param_is_key`. All other paramters of the decorated function are assumed
        to be ``args`` consumed by the lua script.

        By default the decorated method is bound to the :class:`coredis.client.Redis`
        or :class:`coredis.client.RedisCluster` instance that the :class:`Script` instance
        was instantiated with. This may however not be the instance you want to eventually
        execute the method with. For such scenarios the decorated method can accept an additional
        parameter which has the name declared by :paramref:`client_arg`).

        The following example executes the script with the ``client`` instance used
        to register the script. The ``key`` parameter is detected as a key provider
        as it is annotated with the :data:`coredis.typing.KeyT` type, and ``value`` is
        passed to redis as an ``arg``::

            import coredis
            from coredis.typing import KeyT, ValueT
            from typing import List

            client = coredis.Redis()
            @client.register_script("return {KEYS[1], ARGV[1]}").wraps()
            async def echo_key_value(key: KeyT, value: ValueT) -> List[ValueT]: ...

            k, v = await echo_key_value("co", "redis")
            # (b"co", b"redis")

        Alternatively, the following example builds a class method that requires
        the ``client`` to be passed in explicitly::

            from coredis import Redis
            from coredis.commands import Script

            class ScriptProvider:
                @classmethod
                @Script(script="return KEYS[1]").wraps(
                    key_spec=["key"],
                    client_arg="client"
                )
                def echo_key(cls, client, key): ...

                @classmethod
                @Script(script="return ARGS[1]").wraps(
                    client_arg="client"
                )
                def echo_arg(cls, client, value): ...

            echoed = await ScriptProvider.echo_key(Redis(), "coredis")
            # b"coredis"
            echoed = await ScriptProvider.echo_value(Redis(), "coredis")
            # b"coredis"

        :param key_spec: list of parameters of the decorated method that will
         be passed as the :paramref:`keys` argument to :meth:`__call__`. If provided
         this parameter takes precedence over using :paramref:`param_is_key` to determine if
         a parameter is a key provider.
        :param param_is_key: a callable that accepts a single argument of type
         :class:`inspect.Parameter` and returns ``True`` if the parameter points
         to a key that should be appended to the :paramref:`__call__.keys` argument
         of :meth:`__call__`. The default implementation marks a parameter as a key
         provider if it is of type :data:`coredis.typing.KeyT` and is only used if
         :paramref:`key_spec` is ``None``.
        :param client_arg: The parameter of the decorator that will contain a client instance
         to be used to execute the script.
        :param runtime_checks: Whether to enable runtime type checking of input arguments
         and return values. (requires :pypi:`beartype`). If :data:`False` the function will
         still get runtime type checking if the environment configuration ``COREDIS_RUNTIME_CHECKS``
         is set - for details see :ref:`handbook/typing:runtime type checking`.
        :param readonly: If ``True`` forces this script to be called with
         :meth:`coredis.Redis.evalsha_ro`

        :return: A function that has a signature mirroring the decorated function.
        """

        def wrapper(func: Callable[P, Awaitable[R]]) -> Callable[P, Awaitable[R]]:
            sig = inspect.signature(func)
            first_arg = list(sig.parameters.keys())[0]
            runtime_check_wrapper = (
                add_runtime_checks if not runtime_checks else safe_beartype
            )
            script_instance = self
            key_params = (
                key_spec
                if key_spec
                else [n for n, p in sig.parameters.items() if param_is_key(p)]
            )
            arg_fetch: Dict[str, Callable[..., Parameters[Any]]] = {
                n: (lambda v: [v])
                if p.kind
                in {
                    inspect.Parameter.POSITIONAL_ONLY,
                    inspect.Parameter.KEYWORD_ONLY,
                    inspect.Parameter.POSITIONAL_OR_KEYWORD,
                }
                else (lambda v: list(itertools.chain.from_iterable(v.items())))
                if p.kind == inspect.Parameter.VAR_KEYWORD
                else lambda v: list(v)
                for n, p in sig.parameters.items()
            }
            if first_arg in {"self", "cls"}:
                arg_fetch.pop(first_arg)
            if client_arg:
                arg_fetch.pop(client_arg)

            def split_args(
                bound_arguments: inspect.BoundArguments,
            ) -> Tuple[
                Parameters[KeyT],
                Parameters[ValueT],
                Optional[coredis.client.Client[AnyStr]],
            ]:
                bound_arguments.apply_defaults()
                arguments = bound_arguments.arguments
                keys: List[KeyT] = []
                args: List[ValueT] = []
                for name in sig.parameters:
                    if name not in arg_fetch:
                        continue
                    value = arg_fetch[name](arguments[name])
                    if name in key_params:
                        keys.extend(value)
                    else:
                        args.extend(value)
                return (
                    keys,
                    args,
                    arguments.get(client_arg) if client_arg else None,
                )

            @runtime_check_wrapper
            @functools.wraps(func)
            async def __inner(
                *args: P.args,
                **kwargs: P.kwargs,
            ) -> R:
                keys, arguments, client = split_args(sig.bind(*args, **kwargs))
                # TODO: atleast lie with a cast.
                #  mypy doesn't like the cast. pyright is ok with it
                return await script_instance(keys, arguments, client, readonly)  # type: ignore

            return __inner

        return wrapper
